##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: ctsmod_genfunctions.R, v18
# General functions for ctsmod. 


#################### 
#################### Setting functions (not methods)
#################### 

# Given dataframe with column names, return line n as a named list.
# This is used to put prior samples into objects.
# Note that we have to pass the entire dataframe in order to extract the names,
# which are not present when the dataframe has only one column.
set.theta <- function(priorSamples, n) {
  temp <- as.list(priorSamples[n,])
  names(temp) <- names(priorSamples)
  return(temp)
}

# Given a named matrix or dataframe (postSamp), return lines n as a named list,
# selecting only those columns which have names given by the character vector selNames.
# Note that we have to pass the entire matrix/dataframe in order to extract the names,
# which are not present when the matrix/dataframe has only one column.
set2.theta <- function(postSamp, selNames, n) {
  temp <- as.list(postSamp[n, selNames])
  names(temp) <- selNames
  return(temp)
}



####################
#################### Read or generate/simulate data sets. Parameters set in functions
#################### All return obsdata, a matrix with Nevents rows and 4 columns c(s, s.sd, y, y.sd)
####################

# Read observational data from an ASCII file.
# 1 event per row. Either with 4 columns c(s, s.sd, y, y.sd) or 3 columns c(s, y, y.sd).
# In the latter case the s.sd column is added with value 0 for all rows.
# File may contain any number of comment lines, which start with '#'. These are ignored.
# obsdataHeader indicates whether a single header line is present (which may contain anything).
# If so, it is skipped.
read.obs.dat <- function(obsdataFname, obsdataHeader=TRUE, verbose=TRUE) {
  if(verbose) {cat("Reading data from file", obsdataFname, "\n")}
  obsdata <- as.matrix(read.table(obsdataFname, skip=ifelse(obsdataHeader, 1, 0), header=FALSE))
  if(!(ncol(obsdata) %in% c(3,4))) {stop("Data file does not have three or four columns")}
  if(ncol(obsdata)==3) {
    obsdata <- cbind(obsdata[,1], 0, obsdata[,2:3])
  }
  # Perform any basic checks on data here
  colnames(obsdata) <- c("s", "s.sd", "y", "y.sd")
  if(verbose) {cat("Data file has", nrow(obsdata), "events\n")}
  return(obsdata)
}


# sim.const: Nevents drawn from a uniform random distribution in time with
# a signal value drawn from a Gaussian distribution with specified mean and sd.
sim.const <- function(simseed=100) {
  cat("Generating simulated data with sim.const()\n")
  set.seed(simseed)
  Nevents <- 50
  offset.mean <- 0
  offset.sd <- 5
  tmin <- 0
  tmax <- 100
  s.sd <- 0
  y.sd <- 1
  t <- sort(runif(Nevents, min=tmin, max=tmax))
  z <- rnorm(Nevents, mean=offset.mean, sd=offset.sd)
  s <- rnorm(Nevents, mean=t, sd=s.sd)
  y <- rnorm(Nevents, mean=z, sd=y.sd)
  return(cbind(s, s.sd, y, y.sd))
}


# sim.linear: Nevents drawn from a uniform random distribution in time
# and a linear model applied to this. No stochastic component, but
# signal measurement noise added and half this sd reported
# (i.e. intentional underestimation of error bars)
sim.linear <- function(simseed=100) {
  cat("Generating simulated data with sim.linear()\n")
  set.seed(simseed)
  Nevents <- 20
  tmin <-    0
  tmax <-  100
  tmid <-   50
  offset <-  0
  slope  <-  1
  s.sd <- 0
  y.sd <- 10
  t <- sort(runif(Nevents, min=tmin, max=tmax))
  z <- slope*(t - tmid) + offset
  s <- rnorm(Nevents, mean=t, sd=s.sd)
  y <- rnorm(Nevents, mean=z, sd=y.sd)
  return(cbind(s, s.sd, y, y.sd/2))
}


# sim.sinusoid: Nevents drawn from a sinusoidal distribution equally spaced in time
# with Gaussian noise of fixed sd added to time, and Gaussian noise of fixed sd added to signal
# Noise reported can differ from noise added
sim.sinusoid <- function(simseed=100) {
  cat("Generating simulated data with sim.sinusoid()\n")
  set.seed(simseed)
  funType <- 0 # =0 for FuncSinusoidZero (true sinusoid, i.e. symmetric about offset),
               # =1 for FuncSinusoid (minimum of function is offset)
  tmin  <- 9
  tmax  <- 537
  Nevents <- 50
  t   <- seq(from=tmin, to=tmax, length.out=Nevents)
  offset  <- 0
  sinamp  <- 316
  freq    <- 1/62
  phase   <- 5.2/(2*pi)
  z.sd <- rep.int(x=15, times=Nevents) # noise added
  y.sd <- rep.int(x=15, times=Nevents) # noise reported
  s.sd <- rep.int(x=0,  times=Nevents)
  z <- offset + (sinamp/2)*(cos(2*pi*(freq*t + phase)) + funType)
  s <- rnorm(Nevents, mean=t, sd=s.sd)
  y <- rnorm(Nevents, mean=z, sd=z.sd) 
  return(cbind(s, s.sd, y, y.sd))
}


# sim.quasisinusoid: a sinusoidal function (as specified by FuncSinusoid or FuncSinusoidZero)
# with a time dependent phase which varies smoothly in time.
# This variable phase give by a local polynomial fit to data drawn from N(0,1) at each time point.
# The sinusoid is uniformly sampled in time with steps tstep.
# and Gaussian noise of fixed sd added to time, and Gaussian noise of fixed sd added to signal
# Noise reported can differ from noise added.
sim.quasisinusoid <- function(simseed=100) {
  cat("Generating simulated data with sim.quasisinusoid()\n")
  library(KernSmooth)
  set.seed(simseed)
  funType <- 0 # =0 for FuncSinusoidZero (true sinusoid, i.e. symmetric about offset),
               # =1 for FuncSinusoid (minimum of function is offset)
  tmin  <- 9
  tmax  <- 537
  tstep <- 1
  t   <- seq(from=tmin, to=tmax, by=tstep)
  Nevents <- length(t)
  offset <- 0
  sinamp <- 316
  freq   <- 1/62
  phase  <- 5.2/(2*pi)
  # vphi is variable phase, calculated at each t.
  # tscale is the smoothing time scale, in the units of t.
  # To avoid end effects in the smoothing, I pad the time (t2)
  # smooth that (perturb) and then chop (vphi)
  tscale <- 25 
  ext <- seq(from=tstep, to=tstep+tscale, by=tstep)
  t2 <- c(tmin-rev(ext), t, tmax+ext)
  perturb <- locpoly(t2, rnorm(t2), bandwidth=tscale, gridsize=length(t2))
  vphi <- perturb$y[which(perturb$x >= tmin & perturb$x <= tmax)] # must have length(t)
  #
  z <- offset + (sinamp/2)*(cos(2*pi*(freq*t + vphi+ phase)) + funType)
  z.sd <- rep.int(x=50, times=Nevents) # noise added
  y.sd <- rep.int(x=50, times=Nevents) # noise reported
  s.sd <- rep.int(x=0,  times=Nevents)
  s <- rnorm(Nevents, mean=t, sd=s.sd)
  y <- rnorm(Nevents, mean=z, sd=z.sd) 
  return(cbind(s, s.sd, y, y.sd))
}


# sim.asfuncquasisinusoid: simulates a quasi-sinusoidal times series in the
# same way that the TSMod1 FuncQuasiSinusoid models them.
# The sinusoid is uniformly sampled in time with steps tstep.
# and Gaussian noise of fixed sd added to time, and Gaussian noise of fixed sd added to signal
# Noise reported can differ from noise added.
sim.asfuncquasisinusoid <- function(simseed=100) {
  cat("Generating simulated data with sim.asfuncquasisinusoid()\n")
  library(fields)
  set.seed(simseed)
  funType <- 0 # =0 for FuncQuasiSinusoidZero (true sinusoid, i.e. symmetric about offset),
               # =1 for FuncQuasiSinusoid (minimum of function is offset)
  tmin  <- 9
  tmax  <- 537
  Nevents <- 100
  t   <- seq(from=tmin, to=tmax, length.out=Nevents)
  offset  <- 0
  sinamp  <- 316
  freq    <- 1/62
  phase   <- 5.2/(2*pi)
  Nfp     <- 4
  fptimes <- seq(from=0, to=550, length.out=Nfp+1)
  fpX <- c(+0.3, 0.4, -0.4, -0.1) # length = Nfp
  #
  fpvec <- c(0, fpX)
  vphi  <- splint(fptimes, fpvec, xgrid=t) 
  z <- offset + (sinamp/2)*(cos(2*pi*(freq*t + vphi + phase)) + funType)
  z.sd <- rep.int(x=15, times=Nevents) # noise added
  y.sd <- rep.int(x=15, times=Nevents) # noise reported
  s.sd <- rep.int(x=0,  times=Nevents)
  s <- rnorm(Nevents, mean=t, sd=s.sd)
  y <- rnorm(Nevents, mean=z, sd=z.sd) 
  return(cbind(s, s.sd, y, y.sd))
}


# sim.OUprocess: Generated the OU process using gen.OUprocess
# Measurement noise is assumed to exist and can be optionally added to data (if addNoise=TRUE)
# Time uncertainties are zero.
# Note that "offset" is the offset in X, and is just added at the very end to the time series.
sim.OUprocess <- function(simseed=100) {
  cat("Generating simulated data with sim.OUprocess()\n")
  set.seed(simseed)
  relax    <-  10
  diffcon  <-  20
  offset   <-  35
  startT   <-   0
  startX   <-   0 # w.r.t the offset
  startW   <-   0 # just needs to be provided for compatibility
  Nevents  <-  50
  finalT   <- 100
  s.sd     <-   0
  y.sd     <-   1.0
  addNoise <- TRUE
  proc <- gen.OUprocess(relax=relax, diffcon=diffcon, startT=startT, startX=startX, startW=startW, finalT=finalT, Nevents=Nevents)
  if(addNoise) {
    y <- rnorm(Nevents, mean=proc$X, sd=y.sd)
  } else {
    y <- proc$X
  }
  return(cbind(proc$T, rep.int(s.sd, Nevents), y+offset, rep.int(y.sd, Nevents)))
}


# sim.IntOUprocess: Generated the integrated OU process using gen.OUprocess
# Currently no measurement noise is added, yet uncertainties are returned.
sim.IntOUprocess <- function(simseed=100) {
  cat("Generating simulated data with sim.IntOUprocess()\n")
  set.seed(simseed)
  relax   <-  1e-6
  diffcon <-  20
  startT  <-   0
  startX  <-   0
  startW  <-   0
  Nevents <-  50
  finalT  <- 100
  s.sd    <-   0 
  y.sd    <-   0 
  proc <- gen.OUprocess(relax=relax, diffcon=diffcon, startT=startT, startX=startX, startW=startW, finalT=finalT, Nevents=Nevents)
  return(cbind(proc$T, rep.int(s.sd, Nevents), proc$W, rep.int(y.sd, Nevents)))
}


# Given the parameters of an OU process (relax, diffcon), plus the initial conditions
# (startT, startX, startW), the final time (finalT), and the number of events (Nevents),
# generate a (random) OU process and its integral at Nevents-1 time points drawn from a 
# uniform random distribution between startT and finalT, with the first point at time=startT
# Return as a three-column matrix: T,X,W 
gen.OUprocess <- function(relax, diffcon, startT, startX, startW, finalT, Nevents) {
  T <- c(startT, sort(runif(n=Nevents-1, min=0, max=finalT)))
  deltaT <- c(0, diff(T, lag=1))
  X <- vector(length=Nevents)
  W <- vector(length=Nevents)
  mu       <- exp(-deltaT/relax)
  sigX     <- sqrt( diffcon*(relax/2)*(1-mu^2) )
  sigWsq   <- diffcon*relax^3*( deltaT/relax -2*(1-mu) + (1/2)*(1-mu^2) )
  kappaXW  <- diffcon*(relax^2/2)*(1-mu)^2
  fatconst <- sqrt(sigWsq - kappaXW^2/sigX^2) # gives NaN for first element, but this is not needed
  n1 <- rnorm(n=Nevents, mean=0, sd=1)
  n2 <- rnorm(n=Nevents, mean=0, sd=1)
  #
  X[1] <- startX
  W[1] <- startW
  for(i in 2:Nevents) {
    X[i] <- X[i-1]*mu[i] + sigX[i]*n1[i]
    W[i] <- W[i-1] + X[i-1]*relax*(1-mu[i]) + fatconst[i]*n2[i] + (kappaXW[i]/sigX[i])*n1[i]
  }
  return( list(T=T, X=X, W=W) )
}


# sim.Wienerprocess: Generate the Wiener process using gen.Wienerprocess
# Measurement noise is assumed to exist and can be optionally added to data (if addNoise=TRUE)
# Time uncertainties are zero.
# Note that "offset" is the offset in X, and is just added at the very end to the time series.
sim.Wienerprocess <- function(simseed=100) {
  cat("Generating simulated data with sim.Wienerprocess()\n")
  set.seed(simseed)
  diffcon  <-   1
  offset   <-   0
  startT   <-   0
  startX   <-   0 # w.r.t the offset
  Nevents  <-  40
  finalT   <- 100
  s.sd     <-   0
  y.sd     <-   0.1
  addNoise <- TRUE
  proc <- gen.Wienerprocess(diffcon=diffcon, startT=startT, startX=startX, finalT=finalT, Nevents=Nevents)
  if(addNoise) {
    y <- rnorm(Nevents, mean=proc$X, sd=y.sd)
  } else {
    y <- proc$X
  }
  return(cbind(proc$T, rep.int(s.sd, Nevents), y+offset, rep.int(y.sd, Nevents)))
}


# Given the parameter of a Wiener process (diffcon), plus the initial conditions
# (startT, startX), the final time (finalT), and the number of events (Nevents),
# generate a (random) OU process and its integral at Nevents-1 time points drawn from a 
# uniform random distribution between startT and finalT, with the first point at time=startT
# Return as a two-column matrix: T,X
gen.Wienerprocess <- function(diffcon, startT, startX, finalT, Nevents) {
  T <- c(startT, sort(runif(n=Nevents-1, min=0, max=finalT)))
  X <- vector(length=Nevents)
  n1 <- rnorm(n=Nevents, mean=0, sd=1)
  X[1] <- startX
  for(i in 2:Nevents) {
    X[i] <- X[i-1] + sqrt(diffcon*(T[i]-T[i-1]))*n1[i]
  }
  return( list(T=T, X=X) )
}

# sim.asfile: Given a data file of standard format with Nevents, create a time series
# with the same time sampling. Add noise to both time and signal using some noise model
# and the uncertainties on individual events.
# (There is a boolean flag, addNoise, set here to optionally bypass adding noise).
# The resulting s.sd and y.sd will be identical to those in the input file.
# If no noise is added, s will also be identical. y is always changed as we are simulating
# a new time series.
# The mean signal is subtracted from the TS (after adding noise).
# Type and parameters of time series, as well as noise models, are all specified here.
# Current set up:
# TSMod:   TSMod1=FuncCombUniformSinusoid, TSMod2=ProbGaussian, (TSMod3 not relevant as times are preserved)
# MeasMod: TimeUniformSignalGaussian     
sim.asfile <- function(obsdataFname, obsdataHeader, simseed=100) {
  cat("Generating simulated data with sim.asfile()\n")
  set.seed(simseed)
  obsdata <- read.obs.dat(obsdataFname, obsdataHeader)
  Nevents <- nrow(obsdata)
  addNoise <- TRUE
  offset   <- 0
  sinamp   <- max(obsdata[,3]) - min(obsdata[,3])
  freq     <- 1/2.7
  phase    <- 0
  theta2sd <- 0
  z <- offset + (sinamp/2)*cos(2*pi*(freq*obsdata[,1] + phase)) + rnorm(Nevents, mean=0, sd=theta2sd)
  if(addNoise) {
    s.width <- sqrt(12)*obsdata[,2]
    s <- runif(Nevents, min=obsdata[,1]-s.width/2, max=obsdata[,1]+s.width/2)
    y <- rnorm(Nevents, mean=z, sd=obsdata[,4])
  } else {
    s <- obsdata[,1]
    y <- z
  }
  y <- y - mean(y)
  return(cbind(s, obsdata[,2], y, obsdata[,4]))

}



####################
#################### Input parameter/data checking utilities
####################

# Given two named vectors (or lists), do check.partial.match().
# If passed, return second vector/list.
# The first argument must be defined; the second may be NULL
assign.partial <- function(reflist, testlist) {
  if(check.partial.match(reflist, testlist)) {
    return(testlist)
  } else {
    return(NULL)
  }
}


# Given two named vectors (or lists), check that all of the names in the latter are in the former.
# If so, return TRUE, otherwise given an error.
# The first argument must be defined.
# The second may be NULL, in which case TRUE is returned ("nothing" is part of something).
check.partial.match <- function(reflist, testlist) {
  if(is.null(reflist))  {stop("First argument is NULL")}
  if(is.null(testlist)) {
    return(TRUE)
  }
  if(any(!is.element(names(testlist), names(reflist)))) {
    stop("Not all of the names in the second list are in the first list")
  }
  return(TRUE)
}


# Given two named vectors (or lists), do check.full.match().
# If passed, return the second vector/list
assign.full <- function(tobefilled, newvals) {
  if(check.full.match(tobefilled, newvals)) {
    return(newvals)
  } else {
    return(NULL)
  }
}


# Given two named vectors (or lists), check that neither is NULL and that the names
# in the lists are identical (no duplicates, but don't have to be in same order).
# If okay, return TRUE, otherwise return stop. Note that an empty list is not a NULL list.
check.full.match <- function(tobefilled, newvals) {
  if(is.null(names(tobefilled))) {stop("Attempt to assign to non-existent list in object (first argument)")}
  if(is.null(names(newvals))) {stop("Attempt to assign non-existent list (second argument)")}
  if(!setequal(names(tobefilled), names(newvals))) {stop("Names in lists do not agree")}
  if(any(duplicated(names(tobefilled)))) {stop("Names duplicated in list in object (first argument)")}
  if(any(duplicated(names(newvals)))) {stop("Names duplicated in list (second argument)")}
  return(TRUE)
}



####################
#################### Misc. functions
####################

# Assuming that a process will run for Ntot iterations, and has already completed
# n since starting at startTime, return the time elapsed and the total expected duration.
est.time <- function(startTime, n, Ntot) {
  nowTime <- Sys.time()
  elapsedTime <- c(nowTime - startTime)
  elapsedTimeUnits <- units(nowTime - startTime)
  # The following may look bizarre, but this exact format (and parameter order) is required to
  # take advantage of the data types' automatic setting of appropriate time units
  temp <- startTime + (nowTime - startTime)*Ntot/n - startTime
  expectedDuration <- c(temp)
  expectedDurationUnits <- units(temp)
  return( list(elapsedTime=elapsedTime, elapsedTimeUnits=elapsedTimeUnits, expectedDuration=expectedDuration, expectedDurationUnits=expectedDurationUnits) )
}

# Print the time elapsed since the given startTime, which was a call to Sys.time())
print.elapsed.time <- function(initTime) {
  totalTime <- c(Sys.time() - initTime)
  totalTimeUnits <- units(Sys.time() - initTime)
  cat("Time elapsed:", noquote( formatC(totalTime,  format="f", width=5, digits=2, flag="") ), totalTimeUnits, "\n" )
  return(0)
}
