##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: kfoldCV.R, v22
# k-fold Cross Validation. Called by run_ctsmod.R

library(parallel) 

# Divide the complete data set (the events) into Npart disjoint partitions. Given partition k, sample the
# posterior using all the other events (i.e. those not in k). This is done by sample.posterior().
# At each of these samples, calculate the (single) likelihood for the data in partition k.
# The (log10 of the) mean over these samples is called logLikePart (and is also returned - see below).
# Do this for all k=1:Npart partitions.
# For each partition contruct the 2D matrix with one row per sample, with 2+p columns:
# 1:  log10 prior PDF   "logPrior"
# 2:  log10 likelihood  "logLike"
# 3+: the p-dimensional vector of parameters sampled from the posterior, in order
#     thetaFlag[thetaFlag]
# -> This is just the matrix "postSamp" in the list returned by sample.posterior()
# but the likelihoods replaced by the likelihood on the partition k.
# Combine these Npart matrices into a 3D array.
# Return a list with four named elements:
# 1. this 3D array, multiPostSamp[n,i,k] of size Nsamp * (2+p) * Npart
#    Nsamp is the number of rows returned by sample.posterior(). This equals Npostsamp
#    if it's using Metropolis or PT, but equals Npostsamp*Nwalker if it's using emcee
# 2. postSampNames, a character list of the names of the (2+p) columns in each multiPostSamp[,,k]
# 3. logLikePart, a vector of size Npart, giving the log10 of the mean likelihood for each partition,
#    i.e. mean of multiPostSamp[n,i,k] over n, for each k.
# 4. partInd, a list containing Npart vectors, each containing the indices of the events in each partition
#
# Note that do.part() and check.res() both use imlicit passing of variable
# defined in the environment they are defined in, i.e. kfold.cv()
kfold.cv <- function(Npart, obsdata, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, theta1Flag, theta2Flag, theta3Flag, theta1Fixed, theta2Fixed, theta3Fixed, tzlim, integopt, theta1Init, theta2Init, theta3Init, sampleSD, sampleCor, NburnIts, Npostsamp, Nchain, swapInt, Nwalker, Ncores, posteriorSeed, verbose) {

  # Function to do posterior sampling using all obsdata except k^th partition initializing RNG with seed
  do.part <- function(k, seed) {    
    selInd <- setdiff(1:nrow(obsdata), partInd[[k]]) # this is sorted
    verboseP <- ifelse(k==1, verbose, Inf) # only print in sample.posterior for first partition
    if(is.finite(verboseP)) {
      cat("Will show sampling progress for partition 1 only\n")
    }
    postSamp <- sample.posterior(obsdata, selInd, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, theta1Flag, theta2Flag, theta3Flag, theta1Fixed, theta2Fixed, theta3Fixed, tzlim, integopt, theta1Init, theta2Init, theta3Init, sampleSD, sampleCor, NburnIts, Npostsamp, Nchain, swapInt, Nwalker, seed, verboseP)$postSamp # 2D array Nsamp x Ncol (Ncol = 2 + no. adjustable parameters)

    if(is.finite(verboseP)) {
      cat("For partition", k, ": sample.posterior() complete,\n  calculating likelihoods on complementary data\n")
    }
    # Calculate likelihood for k^th partition at each posterior sample and
    # use this to *replace* likelihood in postSamp. Fixed parameters set outside loop.
    # Loop speed dominated by likelihood calculation, so no gain in vectorizing it.
    for(n in 1:nrow(postSamp)) {
      TSMod1@theta1[names(which(theta1Flag))] <- postSamp[n,names(which(theta1Flag))]
      TSMod2@theta2[names(which(theta2Flag))] <- postSamp[n,names(which(theta2Flag))]
      TSMod3@theta3[names(which(theta3Flag))] <- postSamp[n,names(which(theta3Flag))]
      postSamp[n,2] <- calc.like(obsdata, partInd[[k]], TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n) # log10
    }
    if(is.finite(verbose)) {
      est <- est.time(startTime, k, Npart)
      sdump1 <- noquote( formatC(k,        format="d", digits=3, flag="") )
      sdump2 <- noquote( formatC(Npart,    format="g", digits=3, flag="") )
      sdump6 <- noquote( formatC(est$elapsedTime,      format="f", width=5, digits=2, flag="") )
      sdump7 <- noquote( formatC(est$expectedDuration, format="f", width=5, digits=2, flag="") )
      cat("Completed partition", sdump1, "of", sdump2, " Elapsed:", sdump6, est$elapsedTimeUnits, " Expected:", sdump7, est$expectedDurationUnits, "\n")
    }
    return(postSamp) # 2D array Nsamp x Ncol
  }

  # Function to kill all jobs if any job - do.part() execution - has returned an error
  check.res <- function(multiPostSamp) { # multiPostSamp is list of results so far
    sendKill <- FALSE
    for(k in 1:Npart) { # check whether need to kill
      if(inherits(multiPostSamp[[k]], 'try-error')) {
        sendKill <- TRUE
      }
    }
    if(sendKill) { # then kill all
      for(k in 1:Npart) {
        cat("check.res killing job", k, "with PID", job[[k]]$pid, "\n") 
        tools::pskill(job[[k]]$pid)
      }
    }
  }

  # Assign the fixed parameters. Consistency of names is assumed (was checked in calling function).
  # Note that if a parameter is listed in thetaFixed but has been defined as adjustable in thetaFlag,
  # then the latter takes precedence within calc.post.mcmc().
  TSMod1@theta1[names(theta1Fixed)] <- theta1Fixed # Note: names(theta1Fixed) = names(which(!theta1Flag))
  TSMod2@theta2[names(theta2Fixed)] <- theta2Fixed
  TSMod3@theta3[names(theta3Fixed)] <- theta3Fixed

  # Define partitions
  set.seed(posteriorSeed)
  partInd <- define.partitions(Nind=nrow(obsdata), Npart=Npart)

  # Sample posterior for partitions in parallel
  # Does all Npart simultaneously, using as many cores as are available (up to Npart)
# The following allocates cores (seen in top) but they are not used by mcparallel
# Could in principle place the job loop inside a loops which goes Ncore/Npart times
#  if(Ncores>0) { 
#    cl <- makeForkCluster(nnodes=Ncores) 
#  } else { 
#    cl <- makeForkCluster(nnodes=detectCores()) # allocate as many cores as are available
#  }
#  setDefaultCluster(cl)
#  cat("Started ", Npart, "-fold CV using ", length(cl), " (possibly logical) cores\n", sep="")
  startTime <- Sys.time()
  job <- vector("list", length=Npart)
  for(k in 1:Npart) {
    job[[k]]<- mcparallel(do.part(k=k, seed=posteriorSeed+k))
  }
  # any error thrown by do.part() is immediately sent to check.res()
  multiPostSamp <- mccollect(job, intermediate=check.res) # list of length Npart, each containing postSamp 2D matrix,
                                                          # in order 1:Npart, regardless of how long jobs took
#  stopCluster(cl)
  cat("Collected results from", length(multiPostSamp), "parallel jobs\n")

  # Check results and trap some errors
  if(length(multiPostSamp)!=Npart) stop("kfold.cv(): Not all partitions were returned. Cause of error unknown.")
  for(k in 1:Npart) { # cannot access list with vector of indices, so just report first try-error
    if(inherits(multiPostSamp[[k]], 'try-error')) {
      stop("kfold.cv(): Partition ", k, " in multiPostSamp is a try-error.")
    }
  }
  # trap bug1 (see section D.10 of TN)
  failPart <- sapply(multiPostSamp, is.null)
  if(any(failPart)) {
    stop("kfoldcv(): Partition(s) ", paste(which(failPart), collapse=" "), " in multiPostSamp is/are NULL.\n  Cause of error unknown. See section 'bugs' in the technical note")
  }
  
  # Post-process results
  Nsamp <- nrow(multiPostSamp[[1]])
  Ncol  <- 2+length(which(thetaFlag))
  postSampNames <- colnames(multiPostSamp[[1]])
  multiPostSamp <- array( unlist(multiPostSamp), dim=c(Nsamp, Ncol, Npart) ) # convert to 3D array
  #logLikePart <- log10( apply(multiPostSamp, 3, function(x){mean(10^x[,2])} )) # no hfac
  logLikePart <- apply(multiPostSamp, 3, function(x) { # use of hfac could cause problems. Always check results!
    hfac <- -1*median(x[,2])
    log10(mean(10^(x[,2]+hfac))) - hfac
  } ) # x is 2D postSamp matrix for one partition. Returns log10 likelihood for each partition
  
  return( list(multiPostSamp=multiPostSamp, logLikePart=logLikePart, partInd=partInd, postSampNames=postSampNames) ) 

}


# Starting with a vector 1:Nind, assign the elements randomly
# without replacement into Npart partitions of which all but one
# have size partSize = Nind %/% Npart, the other having size
# partSize + Nind %% Npart (which is equal or larger).
# Thus all elements are assigned (and each assigned once).
# E.g. if Nind=22, Npart=4, then three partitions have size 5, and
# one has size 7.
# The partitions are returned as a list of length Npart, with each
# element a vector of the indices (each vector sorted, smallest first).
define.partitions <- function(Nind, Npart) {
  ind <- sample.int(n=Nind)
  partSize <- Nind %/% Npart
  partInd <- NULL
  for(k in 0:(Npart-1)) {
    if(k==(Npart-1)) { # final partition is generally larger
      Len <- partSize + Nind %% Npart 
    } else {
      Len <- partSize
    }
    partInd <- c( partInd, list(sort(ind[ (k*partSize+1):(k*partSize+Len) ])) )
  }
  return(partInd)
}
