##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: mcmc.R, v23
# MCMC algorithms for sampling an arbitrary function. Used by sample_posterior()

library(mvtnorm)

# These MCMC algorithms are quite general. Common features:
#
# The algorithms sample a real vector of variables, thetaInit, from an
# unnormalized posterior probablity density function func.
# "func" must return a two-element vector, the logPrior and logLike
# (log base 10), the sum of which is taken to be the log of the
# density function.  (Either element may be -Inf, but +Inf will cause
# problems - this should never occur if both elements really are a
# logPDF.)
# The sampling method is permitted to produce proposals which lie
# outside the "range" of the parameters: these need to be rejected by
# func returning -Inf in one or both of its elements (typically it
# will be the first, the logPrior).
# The initial values are provided by thetaInit (except for emcee,
# where they are derived from thetaInit and sampleSD) and these must
# give rise to finite values of func().
# The data and other variables needed by func() are passed into these
# functions using the ... parameter.
# If func() is a function of x, but we apply the MCMC to some transformed
# variables, theta=y(x), then we need to take into account the Jacobian, 
# J = |{dx/dy}|. "jacobian" is a function which must return this (a scalar) given
# theta and any other parameters required (passed in/out using ...). Note that
# internal to these MCMC algorithms the parameters are of course the transformed ones,
# theta (and not x; func will need to take care of any inverse transformations.)
# Diagnostics are printed very verbose^th sample: sample number,
# acceptance rate so far, clock time elapsed, and expected clock time
# to completion.
# All other function variables and what is returned are explained for
# the individual functions.
#
# In terms of ctsmod, these algorithms work on the transformed
# variables. func() provides the inverse transformation
# needed to calculate the prior/posterior. jacobian() uses the
# transformed parameter directly.
#
# Note on internal workings:
#
# The initial values of the parameters must be such that func() and
# jacobian() are finite. Provided these conditions are met, and neither
# func() nor jacobian ever return +Inf, then logSR will always be defined.
# ... for metrop() and pt.metrop():
# logSR will be -Inf if the prior or likelihood of the proposal are
# zero, and this leads to that proposal being always rejected (as
# runif never returns exactly 0). Therefore funcCur will always be
# finite, and so we'll never get -Inf - -Inf in the evaluation of
# logSR.
# ... for emcee():
# finite numerical precision could potentially make thetaProp NaN (see
# the TN for an explanation) thus giving rise to an error in func(),
# but this is trapped and logSR is set to -Inf, forcing rejection of
# the proposal. Therefore walkerFunc will always be finite, so we'll
# never get -Inf - -Inf in the evaluation of logSR.



# Metropolis algorithm to sample from func().
# A total of Nburnin+Nsamp samples are drawn, of which the last Nsamp are kept.
# Return a Nsamp * (2+Ntheta) matrix (no names), where the columns are
# 1:  log10 prior PDF
# 2:  log10 likelihood
# 3+: Ntheta parameters
# The sampling PDF is a matrix with fixed covariance matrix given by sampleCov.
# (The order of the parameters in thetaInit and sampleCov must match, of course.)
# As this is symmetric, the Hasting factor cancels in the MCMC, leaving the basic Metropolis algorithm.
# If the parameters have been transformed, then we need to supply a function to 
# return the Jacobian (a scalar) of the parameters. (Note that the parameters within
# metrop() are the transformed parameters.)
metrop <- function(func, jacobian=NULL, thetaInit, sampleCov, Nburnin, Nsamp, verbose, ...) {

  if(is.finite(verbose)) {
    cat("Sampling the posterior with Metropolis algorithm\n")
    cat("iteration # of Nburnin + Nsamp acceptRate Elapsed Expected\n")
  }

  # setup
  isDiag <- ifelse(all(sampleCov[upper.tri(sampleCov)]==0), TRUE, FALSE) # TRUE if sampleCov is diagonal
  sampleSD <- sqrt(diag(sampleCov)) # used to do faster random number drawing below
  Ntheta   <- length(thetaInit)
  funcSamp <- matrix(data=NA, nrow=Nsamp, ncol=2+Ntheta) # this will be filled and returned
  if(is.null(jacobian)) {
    jacobian <- function(...) {
      return(1)
    }
  }
  
  # initialize
  thetaCur <- thetaInit
  funcCur  <- func(thetaInit, ...) # log10
  ljacCur  <- log10(jacobian(thetaCur, ...))
  if(!(all(is.finite(funcCur)))) {
    stop("metrop(): Initialization gives non-finite function values")
  }
  if(!is.finite(ljacCur)) {
    stop("metrop(): Initialization gives non-finite Jacobian")
  }
  nAccept  <- 0
  acceptRate <- 0
  startTime=Sys.time()

  # sample
  for(n in 1:(Nburnin+Nsamp)) {

    # Metropolis algorithm. No Hastings factor for symmetric sampling distributions.
    if(isDiag) { # short-cut
      thetaProp <- rnorm(n=Ntheta, mean=thetaCur, sd=sampleSD)
    } else {
      thetaProp <- rmvnorm(n=1, mean=thetaCur, sigma=sampleCov, method="eigen")
    }
    funcProp <- func(thetaProp, ...) # log10
    ljacProp <- log10(jacobian(thetaProp, ...))
    logSR <- sum(funcProp) - sum(funcCur) + ljacProp - ljacCur # log10 of the Metropolis ratio
    if(is.nan(logSR)) {  # should never occur given other traps
      stop("metrop(): NaN produced, probably because sum(funcProp) = sum(funcCur) = -Inf\n       This should only occur if the logPrior was -Inf for both current and proposal\n       which probably indicates a poor initialization\n       funcProp, funcCur = ", funcProp[1], " ", funcProp[2], ", ", funcCur[1], " ", funcCur[2])
    }
    if(logSR>=0 || logSR>log10(runif(1, min=0, max=1))) { # runif will not generate 0 or 1 exactly
      thetaCur   <- thetaProp
      funcCur    <- funcProp
      ljacCur    <- ljacProp
      nAccept    <- nAccept + 1
      acceptRate <- nAccept/n
    }
    if(n>Nburnin) {
      funcSamp[n-Nburnin,1:2] <- funcCur
      funcSamp[n-Nburnin,3:(2+Ntheta)] <- thetaCur
    }

    # diagnostics
    if( is.finite(verbose) && (n%%verbose==0 || n==Nburnin+Nsamp) ) {
      est <- est.time(startTime, n, (Nburnin+Nsamp))
      sdump1 <- noquote( formatC(n,          format="d", digits=5, flag="") )
      sdump2 <- noquote( formatC(Nburnin,    format="g", digits=5, flag="") )
      sdump3 <- noquote( formatC(Nsamp,      format="g", digits=5, flag="") )
      sdump4 <- noquote( formatC(acceptRate, format="f", width=7, digits=4, flag="") )
      sdump6 <- noquote( formatC(est$elapsedTime,      format="f", width=5, digits=2, flag="") )
      sdump7 <- noquote( formatC(est$expectedDuration, format="f", width=5, digits=2, flag="") )
      cat(sdump1, "of", sdump2, "+", sdump3, sdump4, sdump6, est$elapsedTimeUnits, sdump7, est$expectedDurationUnits, "\n")
    }
    
  }

  return(funcSamp)
 
}


# Parallel tempering Metropolis algorithm to sample from func().
# Nchain chains are run in parallel, with transitions made with the usual Metropolis rule (no Hastings term).
# Every N^th iteration (N=swapInt), a pair of chains is chosen at random, and their values swapped with
# a probability determined by the swaping factor (here logSF).
# The basic Metropolis algorithm is exactly as in the function metrop() - see that function for details, also
# of what func must/may return. The function parameters are the same as metrop() are the same here,
# except that we additionally have Nchain and swapInt.
# The beta values of the Nchains are uniformly spaced from 0 to 1, where 1 is the cold chain.
# Return: funcSamp, which is a 3D array funcSamp[sample, chain, theta], of size
# (Nburnin+Nsamp) * Nchain * (2+Ntheta) matrix (no names)
# which is the set of samples for each chain.
# funcSamp[,chain,] is a 2D array with one row per sample and columns
# 1:  log10 prior PDF
# 2:  log10 likelihood
# 3+: Ntheta parameters
# funcSamp[,Nchain,] is essentially the same as what metrop() returns, except in the current
# function we also return the burn-in samples.
# If the parameters have been transformed, then we need to supply a function to 
# return the Jacobian (a scalar) of the parameters for a given chain. 
# (Note that the parameters within pt.metrop() are the transformed parameters.)
pt.metrop <- function(func, jacobian=NULL, thetaInit, sampleCov, Nburnin, Nsamp, Nchain, swapInt, verbose, ...) {

  if(is.finite(verbose)) {
    cat("Sampling the posterior with parallel tempering Metropolis algorithm with", Nchain, "chains\n")
    cat("iteration # of Nburnin + Nsamp acceptRate swapRate (", Nchain-1, " chain pairs) Elapsed Expected\n", sep="")
  }

  # setup
  isDiag <- ifelse(all(sampleCov[upper.tri(sampleCov)]==0), TRUE, FALSE) # TRUE if sampleCov is diagonal
  sampleSD <- sqrt(diag(sampleCov)) # used to do faster random number drawing below
  Ntheta   <- length(thetaInit)
  funcSamp <- array(data=NA, dim=c(Nburnin+Nsamp, Nchain, 2+Ntheta)) # funcSamp[sample, chain, theta]
  nSwap    <- vector("integer", length=Nchain-1) # number and rate of swaps, for chain pair c,c+1
  swapRate <- vector("numeric", length=Nchain-1) # number of swaps / number of possible swaps (whether made or not)
  beta     <- seq(0,1,length.out=Nchain)
  # To instead drop beta=0 chain:
  # beta <- seq(0,1,length.out=Nchain+1)[(1:Nchain)+1]
  if(is.null(jacobian)) {
    jacobian <- function(...) {
      return(1)
    }
  }
  
  # initialize
  thetaCur <- matrix(rep(thetaInit, Nchain), nrow=Nchain, byrow=TRUE) # thetaCur[chain, theta]
  funcCur  <- matrix(rep(func(thetaCur[1,], ...), Nchain), nrow=Nchain, byrow=TRUE) # funcCur[chain, 1:2]
  ljacCur  <- rep(log10(jacobian(thetaCur[1,], ...)), Nchain) # ljacCur[chain]
  if(!(all(is.finite(funcCur)))) {
    stop("pt.metrop(): Initialization gives non-finite function values")
  }
  if(!(all(is.finite(ljacCur)))) {
    stop("pt.metrop(): Initialization gives non-finite Jacobian")
  }
  nAccept    <- 0 # number and rate of MCMC transitions for cold chain  
  acceptRate <- 0
  startTime  <- Sys.time()

  # sample
  for(n in 1:(Nburnin+Nsamp)) {

    for(c in 1:Nchain) {
      
      # Metropolis algorithm. No Hastings factor for symmetric sampling distribution used.
      if(isDiag) { # short-cut
        thetaProp <- rnorm(n=Ntheta, mean=thetaCur[c,], sd=sampleSD)
      } else {
        thetaProp <- rmvnorm(n=1, mean=thetaCur[c,], sigma=sampleCov, method="eigen")
      }
      funcProp <- func(thetaProp, ...) # log10
      ljacProp <- log10(jacobian(thetaProp, ...))
      logSR <- (funcProp[1] + beta[c]*funcProp[2]) - (funcCur[c,1] + beta[c]*funcCur[c,2]) # log10 of the Metropolis ratio
      logSR <- logSR + ljacProp - ljacCur[c]
      # funcProp = -Inf is possible, and would give logSR=-Inf when beta[c]>0 and thus be rejected
      # by logSR test below. But if beta[c]=0 we'll get logSR=NaN, which is manually rejected by:
      logSR <- ifelse(is.nan(logSR), -Inf, logSR)
      # Minor code bug: the following test can never be met, due to the previous line.
      if(is.nan(logSR)) {   # should never occur given other traps
        stop("pt.metrop(): logSR=NaN, probably because sum(funcProp) = sum(funcCur) = -Inf\n       This should only occur if the logPrior was -Inf for both current and proposal\n       which probably indicates a poor initialization\n       funcProp, funcCur = ", funcProp[1], " ", funcProp[2], ", ", funcCur[c,1], " ", funcCur[c,2], "  beta = ", beta[c])
      }
      if(logSR>=0 || logSR>log10(runif(1, min=0, max=1))) { # runif will not generate 0 or 1 exactly
        thetaCur[c,] <- thetaProp
        funcCur[c,]  <- funcProp
        ljacCur[c]   <- ljacProp 
        if(c==Nchain) {
          nAccept    <- nAccept + 1
          acceptRate <- nAccept/n
        }
      }
      
    } # End of chain loop
    
    if(n %% swapInt == 0) { # potentially make a swap
      cand  <- sample.int(Nchain-1, 1) # select a pair of chains: cand indicates the lower member of the pair
      logSF <- (funcCur[cand,1]   + beta[cand+1]*funcCur[cand,2]) - (funcCur[cand,1]   + beta[cand]*funcCur[cand,2]) +
               (funcCur[cand+1,1] + beta[cand]*funcCur[cand+1,2]) - (funcCur[cand+1,1] + beta[cand+1]*funcCur[cand+1,2])
      # log10 of the swap factor. To have been accepted (above), funcCur must be finite, so logSF must be finite
      if(logSF >= 0 || logSF>log10(runif(1, min=0, max=1))) { # swap; runif will not generate 0 or 1 exactly
        thetaTemp <- thetaCur[cand,]
        funcTemp  <- funcCur[cand,]
        ljacTemp  <- ljacCur[cand]
        thetaCur[cand,]   <- thetaCur[cand+1,]
        funcCur[cand,]    <- funcCur[cand+1,]
        ljacCur[cand]     <- ljacCur[cand+1]
        thetaCur[cand+1,] <- thetaTemp
        funcCur[cand+1,]  <- funcTemp
        ljacCur[cand+1]   <- ljacTemp
        nSwap[cand]    <- nSwap[cand] + 1
        swapRate[cand] <- nSwap[cand]/(n %/% swapInt)
      }
    }

    # store (including the burn-in)
    funcSamp[n,,1:2] <- funcCur
    funcSamp[n,,3:(2+Ntheta)] <- thetaCur

    # diagnostics
    if( is.finite(verbose) && (n%%verbose==0 || n==Nburnin+Nsamp) ) {
      est <- est.time(startTime, n, (Nburnin+Nsamp))
      sdump1 <- noquote( formatC(n,          format="d", digits=5, flag="") )
      sdump2 <- noquote( formatC(Nburnin,    format="g", digits=5, flag="") )
      sdump3 <- noquote( formatC(Nsamp,      format="g", digits=5, flag="") )
      sdump4 <- noquote( formatC(acceptRate, format="f", width=7, digits=4, flag="") )
      sdump5 <- noquote( formatC(swapRate,   format="f", width=7, digits=4, flag="") )
      sdump6 <- noquote( formatC(est$elapsedTime,      format="f", width=5, digits=2, flag="") )
      sdump7 <- noquote( formatC(est$expectedDuration, format="f", width=5, digits=2, flag="") )
      cat(sdump1, "of", sdump2, "+", sdump3, sdump4, sdump5, sdump6, est$elapsedTimeUnits, sdump7, est$expectedDurationUnits, "\n")
    }
    
  } 
  
  return(funcSamp)

}


# affine-invariant ensemble sampler (emcee) to sample from func().
# A set of Nwalker walkers are evolved over Nburnin+Nsamp iterations. (Nwalker>1)
# The walkers are initialized by drawing them from a Gaussian distribution with mean thetaInit
# and covariance sampleCov.
# Return: funcSamp, which is a 3D array funcSamp[iteration, walker, theta], of size
# (Nburnin+Nsamp) * Nwalker * (2+Ntheta) matrix (no names)
# which is the set of all iterations (including the burn-in) for all walkers.
# funcSamp[,walker,] is a 2D array with and one row per sample and columns
# 1:  log10 prior PDF
# 2:  log10 likelihood
# 3+: Ntheta parameters
# v20+: Note that the walkers are updated asynchronously, i.e. a walker is updated in an iteration
# immediately. Thus the walker w_j used to update walker w_k may already have been updated in this
# iteration. This is required to maintain detailed balance. Note that Algorithm 2 in arXiv:1202.3665v3
# is actually wrong in this respect, as it claims that w_j is only drawn from those at iteration t.
# If the parameters have been transformed, then we need to supply a function to 
# return the Jacobian (a scalar) of the parameters. (Note that the parameters within
# metrop() are the transformed parameters.)
emcee <- function(func, jacobian=NULL, thetaInit, sampleCov, Nburnin, Nsamp, Nwalker, verbose, ...) {

  if(is.finite(verbose)) {
    cat("Sampling the posterior with emcee algorithm with", Nwalker, "walkers\n")
    cat("iteration # of Nburnin + Nsamp acceptRate Elapsed Expected\n")
  }

  # setup
  acon <- 2 # if acceptance rate is too low, increase it by lowering acon (and vice versa)
  Ntheta <- length(thetaInit)
  walker <- matrix(data=NA, nrow=Nwalker, ncol=Ntheta) # theta parameters of the walkers
  walkerFunc <- matrix(data=NA, nrow=Nwalker, ncol=2)  # function values of the walkers
  walkerLJac <- vector(mode="numeric", length=Nwalker) # log10(Jacobian) of the walkers
  funcSamp <- array(data=NA, dim=c(Nsamp+Nburnin, Nwalker, 2+Ntheta)) # this will be filled and returned
  if(is.null(jacobian)) {
    jacobian <- function(...) {
      return(1)
    }
  }
  
  # initialize
  # could actually do walker <- rmvnorm(n=Nwalker, mean=thetaInit, sigma=sampleCov, method="eigen")
  isDiag <- ifelse(all(sampleCov[upper.tri(sampleCov)]==0), TRUE, FALSE) # TRUE if sampleCov is diagonal
  for(w in 1:Nwalker) { 
    if(isDiag) { # done to remove dependency on rmvnorm if not needed
      walker[w,] <- rnorm(n=Ntheta, mean=thetaInit, sd=sqrt(diag(sampleCov))) 
    } else {
      walker[w,] <- rmvnorm(n=1, mean=thetaInit, sigma=sampleCov, method="eigen")
    }
    walkerFunc[w,] <- func(walker[w,], ...) # log10
    walkerLJac[w]  <- log10(jacobian(walker[w,], ...))
  }
  if(!(all(is.finite(walkerFunc)))) {
    stop("emcee(): Initialization gives non-finite function values")
  }
  if(!(all(is.finite(walkerLJac)))) {
    stop("emcee(): Initialization gives non-finite Jacobian")
  }
  nAccept  <- 0
  acceptRate <- 0
  startTime=Sys.time()

  # walk 
  for(n in 1:(Nburnin+Nsamp)) {

    for(w in 1:Nwalker) { 
#   for(w in sample.int(Nwalker)) { # do this instead to randomly permute order of updates
      z <- (1/acon)*(1 + (acon-1)*runif(min=0, max=1, n=1))^2
      sel <- sample(x=setdiff(1:Nwalker, w), size=1) # select another walker
      thetaProp <- walker[sel,] + z*(walker[w,] - walker[sel,]) # proposed step
      if(all(is.finite(thetaProp))) {
        funcProp <- func(thetaProp, ...)
        ljacProp <- log10(jacobian(thetaProp, ...))
        logSR <- (Ntheta-1)*log10(z) + sum(funcProp) - sum(walkerFunc[w,]) + ljacProp - walkerLJac[w] # log10 of selection ratio
      } else {
        warning("emcee(): thetaProp = ", thetaProp, "\n  Not all thetaProp are finite so the proposal is being rejected.\n  Probably the theta values are very large (positive or negative), which is not good behaviour")
        logSR <- -Inf
      }
      if(is.nan(logSR)) { # should never occur given other traps
        stop("emcee(): logSR=NaN at n=", n, "w=", w, "funcProp=", funcProp, "walkerFunc=", walkerFunc)
      }
      if(logSR>=0 || logSR>log10(runif(1, min=0, max=1))) { # runif will not generate 0 or 1 exactly
        walker[w,] <- thetaProp
        walkerFunc[w,] <- funcProp
        walkerLJac[w]  <- ljacProp
        nAccept <- nAccept + 1
      } # otherwise walker is unchanged
      funcSamp[n,w,1:2] <- walkerFunc[w,]
      funcSamp[n,w,3:(2+Ntheta)] <- walker[w,]
    } # End of walker loop
    acceptRate <- nAccept/(n*Nwalker)

    # diagnostics
    if( is.finite(verbose) && (n%%verbose==0 || n==Nburnin+Nsamp) ) {
      est <- est.time(startTime, n, (Nburnin+Nsamp))
      sdump1 <- noquote( formatC(n,          format="d", digits=5, flag="") )
      sdump2 <- noquote( formatC(Nburnin,    format="g", digits=5, flag="") )
      sdump3 <- noquote( formatC(Nsamp,      format="g", digits=5, flag="") )
      sdump4 <- noquote( formatC(acceptRate, format="f", width=7, digits=4, flag="") )
      sdump6 <- noquote( formatC(est$elapsedTime,      format="f", width=5, digits=2, flag="") )
      sdump7 <- noquote( formatC(est$expectedDuration, format="f", width=5, digits=2, flag="") )
      cat(sdump1, "of", sdump2, "+", sdump3, sdump4, sdump6, est$elapsedTimeUnits, sdump7, est$expectedDurationUnits, "\n")
    }
    
  }

  return(funcSamp)

}
