##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: run_ctsmod.R v23
# This is the master program

### Function of this program
#
# Calculate the evidence and/or sample the posterior for one time series model (TSMod) for
# one set of observational data (obsdata).
# obsdata is a matrix of J events. For each event we have c(s, s.sd, y, y.sd)
# TSMod <- c(TSMod1, TSMod2, TSMod3)
# This is done calculating the logLikelihood and logPrior at various combinations of the model
# parameters drawn from the prior distributions for these parameters. Some parameters can be fixed.

### Important points
#
# Warning: you must understand the method and all the parameters before using this program! Read the TN!
# In particular
# * the internal parameters of the event likelihood integration, function calc.event.like() are crucial
# * model parameter values (thetaX, alphaX) must be sensible
# * evidences and likelihoods can come out zero (log then -Inf) due to finite precision. The code
#   uses hfac to avoid this, but if you get weird answers, check this.
# The code makes various checks to avoid inconsistent or missing parameters, but not everything is trapped.

### File dependencies
#
# ctsmod_models.R         Defines the classes and methods for the time series and measurement models
# ctsmod_likefunctions.R  Functions for the likelihood calculations
# ctsmod_genfunctions.R   Various general functions
# calc_evidence.R         Functions for calculating the evidence
# kfoldCV.R               Functions for doing K-fold CV (K MCMC samples of posterior)
# sample_posterior.R      Functions for setting up the sampling of the posterior
# mcmc.R                  Functions for sampling the posterior (general purpose)
# setup_ctsmod.R          Where the user sets up all parameters


### Some implementation/parameter setting notes
#
# Evidence: First the (multidimensional) prior of the adjustable parameters is sampled assuming they are independent.
# This is fast. Then the likelihood is calculated at each parameters combination. This is done using a foreach
# loop (see "parallel processing") and is the time-consuming part. Each likelihood calculation generally involves
# J 2D integrations (over t,z), using a numerical integrator (e.g. cuhre{R2Cuba}). However, some significant
# accelerations are possible (see below). The evidence is calculated as the average of the likelihoods.
#
# Posterior sampling: MCMC is used to sample the posterior. As necessary, the model parameters are
# transformed to an infinite range. MCMC is done in serial (one processor).
#
# Object-orientation:
# The three time series models are TSModX (X=1,2,3), implemented as classes using OOP.
# Model parameters are the slot thetaX, the parameters of the prior PDFs over these parameters are in slot alplaX.
#
# Parallel processing:
# Extensive use is made of parallel processing. See the TN for details.
# In general the loop calculations are not done or returned in sequential order.
# Because of the use of parallel and doRNG, ctsmod should not be run in the Mac R.app (or other GUI)
#
# TSMod2 bypass:
# There is an option to leave out the stochastic part of the TS model, achieved by setting
#   TSMod2 <- new("BypassTSMod2")
# For purposes of internal consistency, this forces:
#   theta2Flag  <- c(dummy=FALSE)
#   theta2Fixed <- list(dummy=NA)
# You obviously should not try to assign alpha parameters for this model.
# This results in theta2$dummy being set to NA for all sample selections, which is used to do the bypass
# in calc.event.like() (necessary, because we don't pass flags to that function), i.e. do just a 1D integration.
#
# TSMod2 as purely stochastic model:
# Setting TSMod2 as "ProbOUprocess", "ProbWienerprocess", "ProbOUprocNaive", "ProbIntOUprocNaive" forces
#   TSMod1 <- new("BypassTSMod1")
#   theta1Flag  <- c(dummy=FALSE)
#   theta1Fixed <- list(dummy=NA)
# which bypasses the use of TSMod1 in calc.like(). You should never manually bypass TSMod1.
#
# Numerical integration:
#
# As shown in the TN, the 2D event likelihood integration can sometimes be approximated to (or is equal to)
# a 1D integration or even a 0D integration.
#
# The 1D integration is used for the TSMod2 bypass.
#
# If TSMod2=ProbGaussian, TSMod3=ProbUniform, and the signal part of MeasMod is also Gaussian
# (quite a common case), then the 2D integration even reduces to a simple Gaussian evaluation ("0D integration").
# This is used (for these models) for all events if negtsd=TRUE, or for individual events if s.sd=0.
#
# If s.sd=y.sd=0 and we are not bypassing TSMod2, then the event likelihood is also calculated with a 0D integration.
#
# No (explicit) integration is done if TSMod2=ProbOUprocess or ProbWienerprocess
# (no other settings are required to achieve this).
#
# The global limits on (t,z) for integration are given by tzlim. These should expand beyond the limits of the data
# points in order to accommodate the uncertainties and finite variance of the stochastic processes.
# Whenever a 1D integration is used, the corresponding limit is ignored. If a 0D integration is used, then of course
# neither limit is applied. In such cases the corresponding limits can be set to "NA".
#
# In either the 1D or 2D integrations, shrinkage can be applied to the integration limits for each data point individually.
# This should increase integration speed and accuracy, but you need to think carefully what affect this may have with
# the particular models used.
# When using TSMod2 in the normal way (i.e. neither bypassed, nor ProbOUprocess/ProbWienerprocess) then you should not
# shrink the z integration limits (because the shrinking is determined only from the data, i.e. no allowance for TSMod2).
# If doing 0D integration, shrink is meaningless, so you can set it to anything ('none' would be most logical).

### Run-time issue with parallel processing
#
# The larger Nevsamp, the longer it takes to actually get going with the foreach loop. This seems to
# be because of memory allocation reasons. There is also a hang at the end, which is longer for larger Nevsamp.

### Outputs
#
# 1. Screen dump
# 2. Entire session written out to the file <outStem>.Robj. Because of parallelization involved, the samples
#    may not always be written into the same order in the objects.

### Use of NULL, NA and NaN
#
# In general, NULL means that a parameter does not exist, whereas NA means it has not been set or is not
# to be used. This latter use of NA carries context-dependent information, such as in the case of the
# TSMod2 bypass (theta2$dummy=NA), or when thetaX parameters cannot be adjusted even in principle (alphaXpar=NA).
# NULL is used in the prototype definitions of the classes TSModX and MeasMod, so a "new" object will
# have all its slots set to NULL. Although NA might seem the more obvious choice here, NULL is used
# because at the point of creation we do not know whether all slots will even exist, e.g.
# in the case that a parameter is fixed in which case it has no prior (so the alphas must be NULL).
# NaN is not explicitly used in the code and will only be encountered if something goes wrong.

### Run-time warnings
#
# The integration with cuhre() or integrate() can produce warnings. 
# The cuhre() warning "Dimension out of range" occurs if the likelihood function is not sampled properly, e.g.
# is very narrow for the given number of evaluations (limited by max.eval). (A few such errors are okay.)
# See the help pages of the relevant functions for more information on warnings produced.

### TSMod and MeasMod parameter definition convention
#
# thetaX is a named list of one or more parameters in the function TSModX (X=1,2,3).
# The parameters of the prior PDF of thetaX are in the list alphaX.
# For each element of thetaX, let's generically call this thetaX$par, there is always a corresponding
# element of the same name in the list alphaX, alphaX$par. If thetaX$par can be adjusted, then alphaX$par
# must be either (a) a list containing the parameters of the prior PDF for thetaX$par, or
# (b) set to NULL if the prior PDF has no parameters.
# If thetaX$par cannot be adjusted even in principle, then alphaX$par=NA.
# This is used below as a consistency check against the user request of which parameters to optimize,
# specified in thetaXFlag.
# (Currently the only PDF with no parameters is the uniform one and so returns a constant: TSMod3@theta3.
# But in principle we could have a non-constant PDF with no parameters too)
# The theta and alpha lists are accessed only by name by the calculations, so the order of the
# parameters in them is irrelevant. (However the screen dump - only - does rely on the order within each
# thetaX prototype being the same as the order assigned by sample.prior.theta() ).

### Use of lists and data frames (internal coding issue)
#
# Lists are widely for 1D vectors of TSModX@thetaX and TSModX@alphaX, in which the names of the elements
# are variable, e.g. for the adjustable or fixed parameters. Data frames would not be suitable here,
# because in order to select specific columns we need to know the names as coding time.
# (Data frames would anyway make little sense for this application as they would only have one row.)
# For the same reason I frequently use matrices with named columns, rather than data frames, for
# storing samples of the adjustable parameters (as which parameters are variable is a run-time decision).

### Parameter name convention (internal coding issue)
#
# All function names start lower case and have a dot in them and generally start with verbs,
# e.g. calc.event.like but there are exceptions, e.g. eval.mod.FuncUniform
# Usual variable names start lower case, with concatenated capital words, e.g. logEventLike
# (although sometimes I just have runs of lower case, e.g. obsdata)
# My-defined classes and objects are generally capitalized, e.g. TSMod1


########## START 

cat("ctsmod\nStarted at", date(), "\n")
initTime <- Sys.time()
library(parallel)
source("utilities.R") # not actually needed; but avoids having to load manually afterwards
source("ctsmod_models.R")
source("ctsmod_likefunctions.R")
source("ctsmod_genfunctions.R")
source("calc_evidence.R")
source("kfoldCV.R")
source("sample_posterior.R")
source("mcmc.R")
# Define and source the ctmod setup file, which contains executable code, not function definitions.
# Must either be defined manually or passed as first command-line argument after --args, e.g.
#   R < run_ctsmod.R --no-save --args name_of_setup_file > outfile &
# Otherwise use default name "setup_ctsmod.R". 
if(!exists("setup")) { # attempt to find in command line arguments
  cargs <- commandArgs(trailingOnly=TRUE)
  if(length(cargs)!=0) {
    setup <- cargs[1]
  } else { # use default
    setup <- "setup_ctsmod.R"
  }
}
cat("Sourcing setup file", setup, "\n")
source(setup)
cat("Completed setup\n")


########## Check/set consistency between user-defined parameters and derived parameters, and check they are sensible

if( (verboseE < 1) || (verboseP < 1) ) {stop("verboseX must be >=1 (use Inf to supress)")}

if(class(TSMod2)=="ProbOUprocess" || class(TSMod2)=="ProbWienerprocess" || class(TSMod2)=="ProbOUprocNaive" || class(TSMod2)=="ProbIntOUprocNaive") {
  TSMod1 <- new("BypassTSMod1")
  theta1Flag <- c(dummy=FALSE)
  theta1Fixed <- list(dummy=NA)
}

if(class(TSMod2)=="ProbOUprocess" || class(TSMod2)=="ProbWienerprocess") {
  cat("Note: Use of ProcOUprocess or ProbWienerprocess forces the signal measurement model\n  to be Gaussian and ignores the time uncertainties\n")
}

if(class(TSMod2)=="BypassTSMod2") {
  theta2Flag <- c(dummy=FALSE)
  theta2Fixed <- list(dummy=NA)
}

# Ensure that all the thetaXFlag assignments are logical and are not NA (i.e. are TRUE or FALSE)
if(any(!is.logical(theta1Flag) | is.na(theta1Flag))) {stop("Elements of theta1Flag must be TRUE or FALSE")}
if(any(!is.logical(theta2Flag) | is.na(theta2Flag))) {stop("Elements of theta2Flag must be TRUE or FALSE")}
if(any(!is.logical(theta3Flag) | is.na(theta3Flag))) {stop("Elements of theta3Flag must be TRUE or FALSE")}

cat("Checking that names in thetaXFlag agree with those in TSModX@thetaX\n")
check.full.match(TSMod1@theta1, theta1Flag)
check.full.match(TSMod2@theta2, theta2Flag)
check.full.match(TSMod3@theta3, theta3Flag)

# Ensure no attempt to adapt parameters which are fixed by definition in the class (indicated by alphaX$Z=NA)
if(length(intersect( names(which(theta1Flag)), names(which(is.na(TSMod1@alpha1))) )) !=0) {
  stop("Request has been made in theta1Flag to adjust a parameter in TSMod1 fixed by definition")
}
if(length(intersect( names(which(theta2Flag)), names(which(is.na(TSMod2@alpha2))) )) !=0) {
  stop("Request has been made in theta2Flag to adjust a parameter in TSMod2 fixed by definition")
}
if(length(intersect( names(which(theta3Flag)), names(which(is.na(TSMod3@alpha3))) )) !=0) {
  stop("Request has been made in theta3Flag to adjust a parameter in TSMod3 fixed by definition")
}
thetaFlag <- c(theta1Flag, theta2Flag, theta3Flag) # do not change this order!

if(any(duplicated(names(thetaFlag)))) {stop("Names in thetaFlag are not unique. (This is a programming error.)")}

if(!any(thetaFlag)) {stop("No parameter set for adjustment, so this is all a bit pointless...")}
# See NOTES under "all fixed parameters" for a relaxation of this

# If there are any fixed parameters (as determined by thetaXFlag), ensure that all are assigned (in thetaXFixed).
# The "if" is necessary, because if there are no fixed parameters, thetaXFixed can validly be NULL.
cat("Checking that all fixed parameters have been assigned\n")
if(any(!theta1Flag)) {check.partial.match(theta1Fixed, theta1Flag[!theta1Flag])}
if(any(!theta2Flag)) {check.partial.match(theta2Fixed, theta2Flag[!theta2Flag])}
if(any(!theta3Flag)) {check.partial.match(theta3Fixed, theta3Flag[!theta3Flag])}

# Ensure tzlim are sensible (will give an error if they are not numerical)
if(all(is.na(tzlim[1:2]))) {
  cat("tzlim:   time limits are NA\n")
} else {
  if(tzlim[1] >= tzlim[2]) {stop("tzlim[1]>=tzlim[2] is unacceptable")}
}
if(all(is.na(tzlim[3:4]))) {
  cat("tzlim: signal limits are NA\n")
} else {
  if(tzlim[3] >= tzlim[4]) {stop("tzlim[3]>=tzlim[4] is unacceptable")}
}

# If using TSMod3=ProbUniform, give a warning if its bounds are narrower that tzlim (provided they're not NA)
# (the first checks are done later anyway, but are necessary for the second checks).
if(class(TSMod3)=="ProbUniform" && !all(is.na(tzlim[1:2]))) {
  if(is.null(theta3Fixed$tmin)) {stop("Using TSMod3=ProbUniform but theta3Fixed$tmin not specified")}
  if(is.null(theta3Fixed$tmax)) {stop("Using TSMod3=ProbUniform but theta3Fixed$tmax not specified")}
  if(tzlim[1] < theta3Fixed$tmin) {warning("Lower time limit (tzlim) is below tmin in TSMod3=ProbUniform", immediate.=TRUE)}
  if(tzlim[2] > theta3Fixed$tmax) {warning("Upper time limit (tzlim) is above tmax in TSMod3=ProbUniform", immediate.=TRUE)}
}

# A check is only made that TSModX@alphaX parameters have been set at the point that they are used:
# If not set in eval.prior.theta, then the prior is set to 1 (parameters are assumed to be fixed).
# If not set in in sample.prior.theta, this is an error leading to a stop.

# Check and group likelihood integration options
if(is.na(shrink)) {stop("shrink not recognized")} 
if(!is.logical(negtsd) || is.na(negtsd)) {stop("negtsd must be either TRUE or FALSE")}
integopt <- list(shrink=shrink, negtsd=negtsd)

# Ensure that Npart is non-negative (wouldn't actually cause a problem, but the user should know)
if(Npart<0) {stop("Negative Npart not sensible. Use 0 to skip kfoldCV")}

# Check that NburnIts and Npostsamp are sensible - latter to be >=2 so that matrices are properly defined
if(!is.finite(NburnIts)  || NburnIts<0)  {stop("NburnIts must be >=0 and finite!")}
if(!is.finite(Npostsamp) || Npostsamp<2) {stop("Npostsamp must be >=2 and finite!")} 

# Check that Nchain and swapInt are sensible
if(!is.finite(Nchain)  || Nchain<0)  {stop("Nchain must be >=0 and finite! 1 will use normal Metropolis")}
if(!is.finite(swapInt) || swapInt<1) {stop("swapInt must be >=1 and finite!")}
if(Nchain>1 && swapInt >= (Npostsamp+NburnIts)) {
  warning("swapInt is greater than Npostsamp+NburnIts, so no swaps will occur", immediate.=TRUE)
}

# If emcee is to be used (i.e. if Nchain=0) then Nwalker must be 2 or more.
# In that case check it is sensible, otherwise check that it's set to zero.
# (as we don't want to use more than one MCMC algorithm)
if(Nchain==0) {
  if(!is.finite(Nwalker) || Nwalker<2) {
    if(Nwalker==0) { # give a more general error message
      stop("Both Nchain and Nwalker are zero, preventing any sampling")
    } else { # presumably we wanted to use emcee but chose a bad value (negative, 1 or infinite)
      stop("Nwalker must be >=2 and finite!")
    }
  }
} else {
  if(Nwalker!=0) {stop("Both Nchain and Nwalker are non-zero, whereas only one of these is permitted")}
}
 
# Check that Ncores is sensible
if(!is.finite(Ncores) || Ncores<0) {
  stop("Ncores must be >=0 and finite! Use 0 to take all available cores")
}


########## Read and check data (or generate simulated data)

# obsdata should be a matrix with one row per event, with the four columns (s, s.sd, y, y.sd)
# (are always accessed by position: do not need to be named)
if(is.na(obssel)) {stop("obssel not recognized: ", obssel)}
obsdata <- switch(obssel, fromFile=read.obs.dat(obsdataFname, obsdataHeader), simAsFile=sim.asfile(obsdataFname, obsdataHeader, simseed), simConst=sim.const(simseed), simLinear=sim.linear(simseed), simSinusoid=sim.sinusoid(simseed), simQuasiSinusoid=sim.quasisinusoid(simseed), simAsFuncQuasiSinusoid=sim.asfuncquasisinusoid(simseed), simOUprocess=sim.OUprocess(simseed), simIntOUprocess=sim.IntOUprocess(simseed), simWienerprocess=sim.Wienerprocess(simseed))
if(is.null(obsdata)) {stop("no function defined for obssel = ", obssel)}

# Ensure that time/signal range in tzlim is wider than time/signal range in obsdata
if(!all(is.na(tzlim[1:2]))) {
  if(min(obsdata[,1]) < tzlim[1]) {stop("Smallest obsdata time is below t.lower in tzlim")}
  if(max(obsdata[,1]) > tzlim[2]) {stop("Largest  obsdata time is above t.upper in tzlim")}
}
if(!all(is.na(tzlim[3:4]))) {
  if(min(obsdata[,3]) < tzlim[3]) {stop("Smallest obsdata signal is below z.lower in tzlim")}
  if(max(obsdata[,3]) > tzlim[4]) {stop("Largest  obsdata signal is above z.upper in tzlim")}
}

# If using OU process ensure that tzero is below all times in data
# This may need to apply to ProbOUprocess and ProbWienerprocess too, but maybe not as they
# force not having time measurement uncertainties
if(class(TSMod2)=="ProbOUprocNaive" || class(TSMod2)=="ProbIntOUprocNaive") {
  if(theta2Fixed$tzero >= min(obsdata[,1])) {stop("When using the Naive OU process model, theta2Fixed$tzero must be strictly less than earliest time in data")}
}


######### Configure warnings

options(warn=1) # forces all warnings to be printed as they occur
if(warnToFile) {
  outWarn <- file(description=paste(outStem, ".warnings", sep=""), open="w")
  sink(file=outWarn, type="message")
}


########## Calculate evidence (via prior sampling)

if(calcEvidence) {
  set.seed(priorSeed)
  evCalc <- calc.evidence()
  hfac <- -1*median(evCalc[,2])
  logEvidence <- log10(mean(10^(evCalc[,2]+hfac))) - hfac
  cat("log(Evidence) =", logEvidence, "\n")
  print.elapsed.time(initTime)
}


########## Sample posterior. Either kfoldCV (Npart>1), or, sampling with all data (samplePosterior==TRUE), or both.

# If doing both, then we call them in parallel using mcparallel() and mccollect()
# To save writing long duplicate calls, use expression()
call.kfold.cv <- expression( kfold.cv(Npart, obsdata, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, theta1Flag, theta2Flag, theta3Flag, theta1Fixed, theta2Fixed, theta3Fixed, tzlim, integopt, theta1Init, theta2Init, theta3Init, sampleSD, sampleCor, NburnIts, Npostsamp, Nchain, swapInt, Nwalker, Ncores, posteriorSeed, verboseP) )
call.sample.posterior <- expression( sample.posterior(obsdata, selInd=1:nrow(obsdata), TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, theta1Flag, theta2Flag, theta3Flag, theta1Fixed, theta2Fixed, theta3Fixed, tzlim, integopt, theta1Init, theta2Init, theta3Init, sampleSD, sampleCor, NburnIts, Npostsamp, Nchain, swapInt, Nwalker, posteriorSeed, verboseP) )

##### Do K-fold CV (possibly in parallel with full posterior sampling) and report K-fold CV likelihood.
##### K-fold CV means sample multiple posteriors and calculate and return likelihood on K disjoint data sets

if(Npart>1) {
  if(Npart>nrow(obsdata)) {
    cat("Setting number partitions in K-fold CV equal to number of events\n")
    Npart <- nrow(obsdata)
  }
  if(samplePosterior) { # then do this in parallel with kfoldCV
    cat("Doing full posterior sampling in parallel with kfoldCV (no screen dumps from the former)\n")
    kfoldCVJob <- mcparallel(eval(call.kfold.cv), name="kfoldCV")
    verboseP.temp <- verboseP
    verboseP <- Inf # don't do screen dumps from sample.posterior() when kfold.cv() is running in parallel
    samplePosteriorJob <- mcparallel(eval(call.sample.posterior), name="samplePosterior")
    verboseP <- verboseP.temp
    mycol <- mccollect(list(kfoldCVJob, samplePosteriorJob))
    kfoldCV     <- mycol$kfoldCV
    postSamp    <- mycol$samplePosterior$postSamp
    postSampAll <- mycol$samplePosterior$postSampAll
    rm(mycol, kfoldCVJob, samplePosteriorJob, verboseP.temp)
  } else {
    cat("Doing kfoldCV (no full posterior sampling)\n")
    kfoldCV <- eval(call.kfold.cv)
  }
  cat("kfoldCV log(likelihood) =", sum(kfoldCV$logLikePart), "\n")
  print.elapsed.time(initTime)
}
         
##### If selected, do full posterior sampling, if not just done in parallel with kfoldCV.

if(samplePosterior && Npart<=1) {
  cat("Doing full posterior sampling (no kfoldCV)\n")
  set.seed(posteriorSeed)
  spRes <- eval(call.sample.posterior)
  postSamp    <- spRes$postSamp
  postSampAll <- spRes$postSampAll
  rm(spRes)
  print.elapsed.time(initTime)
}

# Warn on bug1, described in section D.10 of TN
if(samplePosterior && is.null(postSamp)) warning("postSamp and postSampAll are NULL")

##### If selected (and now done), report results from full posterior sampling,
##### i.e. the posterior-averaged likelihood (and DIC)

if(samplePosterior) {
  hfac <- -1*median(postSamp[,2])
  logLike <- log10(mean(10^(postSamp[,2]+hfac))) - hfac
  cat("log(posterior-averaged likelihood) =", logLike, "\n")
#  dic <- calc.dic(postSamp, obsdata, selInd=1:nrow(obsdata), TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt)
#  cat("DIC=D+2Neff =", dic$dic,  "   D =", dic$dic-2*dic$effNpar, "   Neff =", dic$effNpar, "\n-(D/2)/ln10 =", -1/2*(dic$dic-2*dic$effNpar)/log(10), "(=logLike[averageTheta])\n")
}

##### If full posterior sampling has been done with parallel tempering, then use this to estimate
##### the evidence using thermodynamic integration (see TN)

if(samplePosterior && Nchain>1) {
  cmll <- log(10)*apply(postSampAll[(NburnIts+1):(NburnIts+Npostsamp),,2], 2, mean) # chain mean log (base e) likelihood
  beta <- seq(0,1,length.out=Nchain) # consistency assumed with pt.metrop() in sample_posterior.R
  logPTEv <- 0
  for(c in 2:Nchain) { # trapezoidal integration of cmll over beta
    logPTEv <- logPTEv + 0.5*(beta[c]-beta[c-1])*(cmll[c]+cmll[c-1])
  }
  # A simple histogram average would be the average of cmll, i.e.
  # logPTev <- log(10)*mean(postSampAll[(NburnIts+1):(NburnIts+1e3),,2])
  cat("Thermodynamic integration approximation of log10(evidence) =", logPTEv, "\n")
}


########### Likelihood of the No-model (model with no adjustable parameters, defined in TN)

# Calculate the logLike for the no-model, which is sum log10(Event likelihoods).
# As this model has no adjustable parameters, it equals both the evidence and the posterior-averaged likelihood.
# dunif and dnorm return log_e, which I convert to log_10
if(!is.null(nomodPar)) {
  logLikeNomod <- (1/log(10)) * sum( dunif(x=(nomodPar$tmin+nomodPar$tmax)/2, min=nomodPar$tmin, max=nomodPar$tmax, log=TRUE) + dnorm(x=obsdata[,3], mean=mean(obsdata[,3]), sd=obsdata[,4], log=TRUE) )
  cat("logLike(No-model) =", logLikeNomod, "\n")
} else {
  logLikeNomod <- NA
}


########## TS reconstruction - 16 August 2012

# In progress. Removed to TSreconstruct.R


########## Finish up

if(warnToFile) {
  sink(file=NULL, type="message")
  close(outWarn)
}
save(list=ls(all=TRUE), file=paste(outStem, ".Robj", sep=""))
if(sendEmail) {
  host <- system("echo $HOST", intern=TRUE)
  info <- paste("ctsmod complete:", setup, "finished on", host, " at ", date(), sep=" ")
  sysArg <- paste("echo | mailx -s ' ", info, " ' ", EmailAddress)
  system(sysArg)
}
cat("Finished at", date(), "\n")
print.elapsed.time(initTime)
