##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: sample_posterior.R, v23
# Provides:
# 1. sample.posterior(). Sets up the sampling of the posterior (initialization, transformation),
# selects the MCMC sampler (defined in mcmc.R) and post-processes the results (inverse transformation).
# Returns the samples.
# 2. calc.post.mcmc(). Returns the posterior PDF given model parameters and data
# 3. Some other helper functions
# 4. calc.dic(). Returns the DIC given model parameters and data


# Sample the posterior using rows selInd in obsdata.
# One of three alternative MCMC algorithms is used:
# Metropolis, if Nchain=1 (and Nwalker=0)
# Parallel tempering, if Nchain>1 (and Nwalker=0)
# emcee, if Nwalker>1 (and Nchain=0)
# Returns a two element list with members postSamp and postSampAll.
# postSamp is a Npostsamp * (2+p) named matrix, where the columns are
# 1:  log10 prior PDF   "logPrior"
# 2:  log10 likelihood  "logLike"
# 3+: the p-dimensional vector of parameters sampled from the posterior, in order
#     thetaFlag[thetaFlag], in their proper units (i.e. have been transformed back).
# The rows of postSamp are: for Metropolis, just the samples; for parallel
# tempering, the samples from the cold chain; for emcee, all samples for all walkers,
# ordered in blocks of all walkers for each iteration: walker + Nwalker*(iteration-1).
# postSampAll is not filled by Metropolis, in which case it is NULL. Otherwise it is a
# (NburnIts + Npostsamp) * Nchain  * (2+Ntheta) array for parallel tempering,
#   storing for each chain the matrix postSamp, or
# (NburnIts + Npostsamp) * Nwalker * (2+Ntheta) array for emcee,
#   storing for each walker the matrix postSamp.
# TSModX and MeasMod need to be defined but not filled.
# The function first assigns the fixed and adjustable parameters to TSModX@thetaX.
# I use these models in order to access their transformation methods, as needed by the MCMC.
# After transformation we build the thetaAdjInit vector from the model, which is passed to the MCMC algorithm.
# The samples returned by this are finally inverse transformed (again via the models).
sample.posterior <- function(obsdata, selInd, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, theta1Flag, theta2Flag, theta3Flag, theta1Fixed, theta2Fixed, theta3Fixed, tzlim, integopt, theta1Init, theta2Init, theta3Init, sampleSD, sampleCor, NburnIts, Npostsamp, Nchain, swapInt, Nwalker, seed, verbose) {

  if(is.finite(verbose)) {cat("Started posterior sampling using", length(selInd), "events\n")}
  set.seed(seed)
  
  # Assign the fixed parameters. Consistency of names is assumed (was checked in calling function).
  # Note that if a parameter is listed in thetaFixed but has been defined as adjustable in thetaFlag,
  # then the latter takes precedence within calc.post.mcmc().
  TSMod1@theta1[names(theta1Fixed)] <- theta1Fixed # Note: names(theta1Fixed) = names(which(!theta1Flag))
  TSMod2@theta2[names(theta2Fixed)] <- theta2Fixed
  TSMod3@theta3[names(theta3Fixed)] <- theta3Fixed
  
  # Assign initial values for the adjustable parameters.
  # The "if" is necessary, because if there are no adjustable parameters, thetaXFlag and thetaXInit 
  # are empty or null (which would cause assign.full to stop).
  # assign.full() ensures the two lists in its arguments are not empty/null and are identical.
  # Yet even if identical, they may validly be in a different order.
  # To ensure a correct assignment, I use thetaXInit[temp] rather than thetaXInit:
  # temp gives the names of the adjustable parameters in TSModX@thetaX in the order in thetaXFlag. 
  # Thus thetaXInit[temp] gives the parameters in the same order (needed for the assignment),
  # and this is returned by assign.full() (if the lists are identical, otherwise it stops).
  # If temp contains names not in theta1Init, these appear as NA and that error is caught by assign.full().
  if(is.finite(verbose)) {cat("Assigning initial values to adjustable parameters\n")}
  if(any(theta1Flag)) {
    temp <- names(which(theta1Flag))
    TSMod1@theta1[temp] <- assign.full(theta1Flag[theta1Flag], theta1Init[temp])
  }
  if(any(theta2Flag)) {
    temp <- names(which(theta2Flag))
    TSMod2@theta2[temp] <- assign.full(theta2Flag[theta2Flag], theta2Init[temp])
  }
  if(any(theta3Flag)) {
    temp <- names(which(theta3Flag))
    TSMod3@theta3[temp] <- assign.full(theta3Flag[theta3Flag], theta3Init[temp])
  }
  
  # Transform the (adjustable) parameters
  TSMod1@theta1 <- trans.theta(TSMod1, theta1Flag)
  TSMod2@theta2 <- trans.theta(TSMod2, theta2Flag)
  TSMod3@theta3 <- trans.theta(TSMod3, theta3Flag)
  
  # Create vector of (transformed) initial values, in the same order as the adjustable parameter in thetaFlag
  # Ensure that the names agree in sampleSD agree with this. Then generate the MCMC sampling covariance matrix,
  # ensuring that sampleSD is mapped into the same order as thetaAdjInit.
  thetaAdjInit <- c( TSMod1@theta1[names(which(theta1Flag))], TSMod2@theta2[names(which(theta2Flag))], TSMod3@theta3[names(which(theta3Flag))] )
  if(is.finite(verbose)) {cat("Checking agreement between adjustable parameters and sampling matrix\n")}
  check.full.match(sampleSD, thetaAdjInit)
  temp <- as.numeric(sampleSD[names(thetaAdjInit)]) # puts sampleSD into same order as thetaAdjInit
  sampleCov <- make.covariance.matrix(temp, sampleCor)
  thetaAdjInit <- as.numeric(thetaAdjInit)

  # Check whether the alphas for the adjustable parameters have been set as necessary.
  # This uses the check code in sample.prior, either causing a stop if the check fails
  # otherwise returning empty lists. (This does not incremement the random number generator)
  # This was changed in v15. Old: theta1Fixed  New: assign.partial(TSMod1@theta1, theta1Fixed)
  # to be compatible with calc_evidence.R. Doesn't seem to have changed anything
  sample.prior.theta(TSMod1, theta1Flag, assign.partial(TSMod1@theta1, theta1Fixed), 0)
  sample.prior.theta(TSMod2, theta2Flag, assign.partial(TSMod2@theta2, theta2Fixed), 0)
  sample.prior.theta(TSMod3, theta3Flag, assign.partial(TSMod3@theta3, theta3Fixed), 0)

  # thetaAdjInit is a vector of the adjustable parameters, in which the order is relevant (names are not used):
  # the order must be (has been set to be) the same as the order in thetaFlag.
  # We need to pass initialized TSModX in which the alphas and the fixed thetas have been set.
  # MeasMod can be entirely empty.
  # There are three choices of sampling algorithm: Metropolis, Parallel tempering, emcee
  # only one of which is used according to the values of Nchain and Nwalker (values of which have already
  # been checked/restricted in run_ctsmod.R)
  # metrop() draws samples (thetaAdj) from the Gaussian proposal distribution with covariance sampleCov.
  # pt.metrop() is similar, but uses parallel chains (tempering)
  # emcee() uses EMCEE instead of the Gaussian proposal (which is only used to initialize the walkers).
  # All three algorithms pass the samples generated to func=calc.post.mcmc where they are put into TSModX for
  # calculating the prior and likelihood. n is not used. It is passed for compatibility with calc.like()
  # which is called by calc.post.mcmc()
  
  if(Nchain==1) { 
    postSampAll <- NULL
    postSamp <- metrop(func=calc.post.mcmc, jacobian=calc.jacobian.mcmc, thetaInit=thetaAdjInit, sampleCov=sampleCov, Nburnin=NburnIts, Nsamp=Npostsamp, verbose=verbose, obsdata=obsdata, selInd=selInd, TSMod1=TSMod1, TSMod2=TSMod2, TSMod3=TSMod3, MeasMod=MeasMod, thetaFlag=thetaFlag, tzlim=tzlim, integopt=integopt, n=NA) # does not include the burn-in
  }
  if(Nchain>1) {
    postSampAll <- pt.metrop(func=calc.post.mcmc, jacobian=calc.jacobian.mcmc, thetaInit=thetaAdjInit, sampleCov=sampleCov, Nburnin=NburnIts, Nsamp=Npostsamp, Nchain=Nchain, swapInt=swapInt, verbose=verbose, obsdata=obsdata, selInd=selInd, TSMod1=TSMod1, TSMod2=TSMod2, TSMod3=TSMod3, MeasMod=MeasMod, thetaFlag=thetaFlag, tzlim=tzlim, integopt=integopt, n=NA)
    postSamp <- postSampAll[(NburnIts+1):(NburnIts+Npostsamp),Nchain,] # just the non-burn-in samples for the cold chain
  }
  if(Nchain==0 && Nwalker>=2) {
    postSampAll <- emcee(func=calc.post.mcmc, jacobian=calc.jacobian.mcmc, thetaInit=thetaAdjInit, sampleCov=sampleCov, Nburnin=NburnIts, Nsamp=Npostsamp, Nwalker=Nwalker, verbose=verbose, obsdata=obsdata, selInd=selInd, TSMod1=TSMod1, TSMod2=TSMod2, TSMod3=TSMod3, MeasMod=MeasMod, thetaFlag=thetaFlag, tzlim=tzlim, integopt=integopt, n=NA)
    postSamp <- matrix( aperm(postSampAll[(NburnIts+1):(NburnIts+Npostsamp),,], c(2,1,3) ), nrow=Npostsamp*Nwalker, ncol=dim(postSampAll)[3])
    # postSamp is the non-burn-in samples for all walkers. The iteration number varies more slowly in the rows,
    # i.e. the rows give all walkers for the first iteration, then all for the second iteration, etc.
  }
    
  # Inverse transform the posterior parameter samples, postSamp and postSampAll
  # First columns 3+ of postSamp.
  if(is.finite(verbose)) {cat("MCMC sampling complete. Inverse transforming results\n")}
  postSampNames <- c("logPrior", "logLike", names(which(thetaFlag)) )
  colnames(postSamp) <- postSampNames
  rownames(postSamp) <- c(1:nrow(postSamp))
  p1 <- invtrans.multi(postSamp, TSMod1, "theta1", theta1Flag)
  p2 <- invtrans.multi(postSamp, TSMod2, "theta2", theta2Flag)
  p3 <- invtrans.multi(postSamp, TSMod3, "theta3", theta3Flag)
  # cbind ignores any argument which is NULL. 
  # cbind works if any pX is a vector, but the column name (if any) is then lost. So these are reimposed.
  postSamp <- cbind(postSamp[,1:2], p1, p2, p3) 
  colnames(postSamp) <- postSampNames
  # Now do the same for postSampAll, if it has been filled
  if(!is.null(postSampAll)) {
    for(w in 1:dim(postSampAll)[2]) {
      postSampSel <- postSampAll[,w,]
      colnames(postSampSel) <- postSampNames
      rownames(postSampSel) <- c(1:nrow(postSampSel))
      p1 <- invtrans.multi(postSampSel, TSMod1, "theta1", theta1Flag)
      p2 <- invtrans.multi(postSampSel, TSMod2, "theta2", theta2Flag)
      p3 <- invtrans.multi(postSampSel, TSMod3, "theta3", theta3Flag)
      postSampAll[,w,] <- cbind(postSampSel[,1:2], p1, p2, p3) 
    }
  }

  return(list(postSamp=postSamp, postSampAll=postSampAll))
  
}



# Calculate the logarithm (base 10) of the prior and the likelihood
# (the sum of which is the unnormalized log posterior) for rows selInd
# in obsdata using the values of the adjustable parameters given in
# the vector thetaAdj. This requires evaluation the model functions at
# the events selInd for fixed model parameters.
# Return a two element vector: logPrior, logLike.
# These must both be numeric. A value of -Inf is acceptable, but +Inf,
# NA, NaN, NULL are not (and are trapped).
# The names of these parameters are taken to be names(which(thetaFlag)),
# i.e. those elements of thetaFlag which are TRUE, in that order.
# TSModX@thetaX is filled with thetaAdj in this function,
# whereas the fixed parameters - theta, alpha - must already be assigned within TSModX.
# Note that the adjustable parameter are the transformed ones.
# MeasMod need only be initialized using "new", but nothing need be assigned.
# tzlim and integopt are used in the likelihood calculation - see the relevant function.
# n is used for diagnostics in calc.like().
# The ... is not strictly required because the current call to each MCMC algorithm passes
# in only the union of variables explicitly named in the MCMC definition and the
# variables named here. But it's safer to add this in case we later pass other variables
# into the MCMC which would be used by some other function.
calc.post.mcmc <- function(thetaAdj, obsdata, selInd, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n, ...) {

  # Assign the thetaAdj values to the appropriate parameters in TSModX@thetaX for X=1,2,3.
  # This assignment assumes that the TRUE elements of thetaFlag match 1-to-1 and in order with the
  # elements of thetaAdj, but no other ordering (in particular in TSModX@thetaX) is assumed.
  thetaAdjNames <- names(which(thetaFlag))
  # If ind is empty, nothing is changed
  ind <- which(is.element( thetaAdjNames, names(TSMod1@theta1) ))  
  TSMod1@theta1[thetaAdjNames[ind]] <- thetaAdj[ind]
  TSMod1@theta1 <- invtrans.theta(TSMod1, thetaFlag)
  ind <- which(is.element( thetaAdjNames, names(TSMod2@theta2) ))  
  TSMod2@theta2[thetaAdjNames[ind]] <- thetaAdj[ind]
  TSMod2@theta2 <- invtrans.theta(TSMod2, thetaFlag)
  ind <- which(is.element( thetaAdjNames, names(TSMod3@theta3) ))  
  TSMod3@theta3[thetaAdjNames[ind]] <- thetaAdj[ind]
  TSMod3@theta3 <- invtrans.theta(TSMod3, thetaFlag)
  
  logPrior <- log10( eval.prior.theta(TSMod1) * eval.prior.theta(TSMod2) * eval.prior.theta(TSMod3) )
  if(is.null(logPrior) || is.na(logPrior) || identical(logPrior, +Inf)) { # (also traps NaN) - should never happen
    dump <- unlist(c(TSMod1@theta1, TSMod2@theta2, TSMod3@theta3))
    dump <- noquote( formatC(as.numeric(dump), format="g", width=9, digits=3, flag="") )
    stop("calc.post.mcmc():  Unacceptable value: logPrior=", logPrior, " with theta = ", dump, "\n")
  }
  # logPrior permitted to be -Inf, although if this was because of an incongruity between
  # a fixed parameter and its prior (alpha parameters), that's not good.
  if(is.infinite(logPrior)) { # only if -Inf
    logLike <- -Inf
  } else {
    logLike  <- calc.like(obsdata, selInd, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n) # log base 10
  }
  if(is.null(logLike) || is.na(logLike) || identical(logLike, +Inf)) { # (also traps NaN) - should never happen
    dump <- unlist(c(TSMod1@theta1, TSMod2@theta2, TSMod3@theta3))
    dump <- noquote( formatC(as.numeric(dump), format="g", width=9, digits=3, flag="") )
    stop("calc.post.mcmc():  Unacceptable value: logLike=", logLike, " with theta = ", dump, "\n")
  }
  
  return( c(logPrior, logLike) ) # not normalized (as evidence not calculated, of course)
}

# Calculate the Jacobian corresponding to the various transformations used in 
# the MCMC. As all transformations are 1D, the Jacobian is just the product of 
# individual transformations, so this function returns a scalar (unity if there
# are no transformations). The behaviour is similar to calc.post.mcmc().
# Note that the parameters passed in via TSModX are the transformed ones, 
# which is what the functions jacobian.theta() also assume.
# The ... is here to accept (and ignore) other parameters which are also
# accepted by the MCMC algorithm as ... and therefore passed to the call of
# this function as ..., but are not required here.
calc.jacobian.mcmc <- function(thetaAdj, TSMod1, TSMod2, TSMod3, thetaFlag, ...) {
  
  thetaAdjNames <- names(which(thetaFlag))
  ind <- which(is.element( thetaAdjNames, names(TSMod1@theta1) ))  
  TSMod1@theta1[thetaAdjNames[ind]] <- thetaAdj[ind]
  ind <- which(is.element( thetaAdjNames, names(TSMod2@theta2) ))  
  TSMod2@theta2[thetaAdjNames[ind]] <- thetaAdj[ind]
  ind <- which(is.element( thetaAdjNames, names(TSMod3@theta3) ))  
  TSMod3@theta3[thetaAdjNames[ind]] <- thetaAdj[ind]
  
  jac <- prod(jacobian.theta(TSMod1, thetaFlag), jacobian.theta(TSMod2, thetaFlag), jacobian.theta(TSMod3, thetaFlag))
  # Provided each variable is within its proper range, the Jacobian must be zero or positive.
  # It's possible that if variable is at or very close to its extreme value, 
  # then jac=+Inf, which would then need to be trapped.
  if(is.null(jac) || is.na(jac) || identical(jac, +Inf) || jac<0) { # (also traps NaN)
    stop("calc.jacobian.mcmc(): Bogus value of Jacobian: ", jac, "\n")
  }
  return(jac)
}

# Given a named 2D matrix of thetaX parameter samples, samp, (columns are parameters, rows are samples)
# with names from the model TSModX, apply to all the adjustable parameters (as given by thetaXFlag) 
# in all rows the inverse transformation specified by the method invtrans.theta() appropriate for each
# model/parameter.
# Note that thetaX must be supplied as a character string, e.g. "theta2" not theta2
# Return a 2D matrix containing just the columns of samp corresponding to the adjustable parameters
# for model TSModX, inverse transformed. If there no columns in samp corresponding to thetaXFlag==TRUE,
# (i.e. only fixed parameters for that model), return NULL.
# The ordering of the columns in what is returned is the ordering of thetaXFlag (for those which are TRUE),
# which is not necessarily the ordering in the input (samp) (although so far in ctsmod I call this function
# with samp in that order too).
# This function is complex, because of the difficulty in vectorizing invtrans.theta().
invtrans.multi <- function(samp, TSModX, thetaX, thetaXFlag) {

  # Assign the named list samprow to TSModX@thetaX
  # The (un)listing was the result of trial and error...
  assign.theta <- function(TSModX, thetaX, samprow) {
    slot(TSModX, thetaX) <- as.list(unlist(samprow)) # replace the whole slot thetaX
    return(TSModX)
  }

  if(all(!thetaXFlag)) { # if no parameters are adjustable
    return(NULL)
  }

  Nsamp <- nrow(samp)
  # The explicit matrix conversion and column naming is necessary in case there is only one adjustable parameter
  relSamp <- as.matrix(samp[,names(which(thetaXFlag))]) 
  colnames(relSamp) <- names(which(thetaXFlag))
  big <- vector(mode="list", length=Nsamp)
  big <- mapply(FUN='<-', big, MoreArgs=list(TSModX)) # list of Nsamp TSModX
  # write each row of relSamp to each big[[n]]@thetaX. This contains only the adjustable parameters.
  # the apply() is necessary to make each row of relSamp a single list
  big <- mapply(FUN=assign.theta, big, apply(relSamp, 1, list), MoreArgs=list(thetaX=thetaX))
  # temp is normally a p x Nsamp matrix of mode list, where p is no. of adjustable parameters
  # invtrans.theta() returns NULL for any parameters in thetaXFlag which are not in big[[i]]@thetaX, i.e. for
  # the fixed parameters. Hence temp gets a row of NULLs with the rowname "NA" corresponding to such parmaeters.
  # The order of the parameters (rows) is that given in the definition of TSModX, not the one in thetaXFlag!
  # (If thetaXFlag has only one element, then temp is returned in a different form, but the subsequent
  #  can cope with that.)
  temp <- mapply(FUN=invtrans.theta, big, MoreArgs=list(thetaXFlag))
  relSamp <- matrix(unlist(temp), nrow=Nsamp, byrow=TRUE) # ignores any NULL rows
  if(ncol(relSamp)>1) {                                   # then temp has >1 row and hence rownames is defined
    colnames(relSamp) <- setdiff(rownames(temp), NA)      # drop out the name of any NULL row
    return( relSamp[,names(which(thetaXFlag))] )          # puts into the order of thetaXFlag
  } else {                                                # there is only one adjustable parameter; return unnamed vector
    return(relSamp)
  }
}



# Return a covariance matrix given a vector of the standard deviations
# and the global (i.e. scalar) correlation coefficient.
# This is inefficient, but matrix is very small and only done once.
make.covariance.matrix <- function(sampleSD, sampleCor) {
  Len <- length(sampleSD)
  covMatrix <- matrix(nrow=Len, ncol=Len)
  if(abs(sampleCor)>1) {stop("|sampleCor| > 1")}
  for(i in 1:(Len)) {
    for(j in 1:(Len)) {
      if(i==j) {
        covMatrix[i,j] <- sampleSD[i]^2
      } else {
        covMatrix[i,j] <- sampleCor*sampleSD[i]*sampleSD[j]
      }
    }
  }
  return(covMatrix)
}



# Calculate and return DIC and corresponding effective number of parameters
# using the posterior samples and corresponding likelihoods already calculated (postSamp)
# for the subset of events (rows) selInd in obsdata. The data must be passed again
# because we have to calculate the likelihood at the average parameters. It is up to the
# user to ensure that these data are appropriate for the posterior samples and likelihoods
# you pass in! (Typically one will have used the whole data set for all of this.)
# Returns named list containing DIC and effective number of parameters.
calc.dic <- function(postSamp, obsdata, selInd, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt) {

  # Calculate the mean of the deviance over the posterior samples 
  meanDev <- -2*log(10)*mean(postSamp[,2])

  # Calculate the deviance at the mean of the posterior samples
  TSMod1@theta1[names(theta1Fixed)] <- theta1Fixed
  TSMod2@theta2[names(theta2Fixed)] <- theta2Fixed
  TSMod3@theta3[names(theta3Fixed)] <- theta3Fixed
  thetaMean <- apply(X=as.matrix(postSamp[,3:ncol(postSamp)]), MARGIN=2, FUN=mean) 
  thetaAdjNames <- names(which(thetaFlag))
  # If ind is empty, nothing is changed
  ind <- which(is.element( thetaAdjNames, names(TSMod1@theta1) ))  
  TSMod1@theta1[thetaAdjNames[ind]] <- thetaMean[ind]
  ind <- which(is.element( thetaAdjNames, names(TSMod2@theta2) ))  
  TSMod2@theta2[thetaAdjNames[ind]] <- thetaMean[ind]
  ind <- which(is.element( thetaAdjNames, names(TSMod3@theta3) ))  
  TSMod3@theta3[thetaAdjNames[ind]] <- thetaMean[ind]
  devMean <- -2*log(10)*calc.like(obsdata, selInd, TSMod1, TSMod2, TSMod3, MeasMod, thetaFlag, tzlim, integopt, n=NA)

  effNpar <- meanDev - devMean
  dic <- devMean + 2*effNpar
  return( list(dic=dic, effNpar=effNpar) )  
}
