##### ctsmod software, copyright 2012 Coryn A.L. Bailer-Jones
##### http://www.mpia.de/homes/calj
##### This software is released under the Creative Commons 
##### Attribution-NonCommercial-ShareAlike 3.0 Unported (CC BY-NC-SA 3.0) license.
##### See the file "license" for more details

# This file: utilities.R, v22
# Collection of functions for analysing results, plotting etc.
# Intended to be sourced

# Functions for converting specific file formats into standard four-column input file
# are in the corresponding NOTES files in the results directories, e.g. results/bdvar/bdvar_notes

#source("ctsmod_models.R")
library(gplots)
library(fields)

##### Plot functions

# General gamma plotting function
# yzero=TRUE forces ymin=0, otherwise it's calculated from the data
# oplot=TRUE and fname=NA, will just overplot in red on whatever is open
# mean=shape*scale, variance=shape*(scale^2)
# scale=var/mean, shape=mean^2/var
plot.gamma <- function(mean, sd, scale=sd^2/mean, shape=mean^2/sd^2, xmin=0, xmax=10, yzero=FALSE, fname=NA, oplot=FALSE) {
  if(!is.na(fname)) {pdf(fname, width=7, height=4)}
  x <- seq(from=xmin, to=xmax, length.out=1000)
  y <- dgamma(x, shape=shape, scale=scale)
  ymin <- ifelse(yzero, 0, min(y))
  yrange <- c(ymin, range(y, finite=TRUE)[2])
  if(oplot && is.na(fname)) {
    lines(x, y, col="red")
  } else {
    plot(x, y, type="l", ylim=yrange, ylab="dgamma", main=paste("Gamma distribution: scale =", scale, "  shape =", shape) )
    grid()
  }
  if(max(y)==Inf) {text(0,yrange[2],"Inf")}
  if(!is.na(fname)) {dev.off()}
}

# General sigmoid plotting function
plot.sigmoid <- function(sigamp, lambda, tzero, tmin=0, tmax=100, ylim=c(0,sigamp)) {
  t <- seq(from=tmin, to=tmax, length.out=1000)
  y <- sigamp / ( 1+exp(-(t-tzero)/lambda) )
  plot(t, y, type="l", ylab="sigmoid", ylim=ylim, main=paste("sigamp =", sigamp, "  lambda =", lambda, "  tzero =", tzero) )
}



##### Plot obsdata frame

# Function for plotting a data set and its uncertainties (sd)
# obsdata should be a matrix with one row per event, with four columns (s, s.sd, y, y.sd)
# If ylim=NULL then calculate y limits from data (but use ymin=0 if yzero=TRUE)
# If no filename is supplied then it plots to the screen
# Requires library(gplots)
plot.data <- function(obsdata=obsdata, fname=NA, yzero=FALSE, ylim=NULL) {
  if(!is.na(fname)) {
    pdf(fname, width=6, height=3)
  }
  par(mgp=c(2.0,0.8,0), mar=c(3.5,4.0,1.0,1.0))
  xlim <- range(c(obsdata[,1]+obsdata[,2], obsdata[,1]-obsdata[,2]))
  if(is.null(ylim)) {
    ymin <- ifelse(yzero, 0, min(obsdata[,3]-obsdata[,4]))
    ylim <- c(ymin, max(obsdata[,3]+obsdata[,4]))
  }
  plot(obsdata[,1], obsdata[,3], xlim=xlim, ylim=ylim, pch=18, xlab="measured time s", ylab="measured signal y")
  if(any(obsdata[,2]!=0)) {
    plotCI(obsdata[,1], obsdata[,3], uiw=obsdata[,2], err="x", type="n", gap=0, sfrac=0.01, add=TRUE)
  }
  if(any(obsdata[,4]!=0)) {
    plotCI(obsdata[,1], obsdata[,3], uiw=obsdata[,4], err="y", type="n", gap=0, sfrac=0.01, add=TRUE)
  }
  if(!is.na(fname)) {
    dev.off()
  }
}



##### Calculate statistics from postSamp, kfoldCV$multiPostSamp, even evCalc

# Many of the following functions are passed a matrix of samples, samp
# which has columns
# 1: logPrior (base 10)
# 2: logLike  (base 10)
# 3: the parameters
# samp can be postSamp (and in some cases kfoldCV$multiPostSamp[,,k]) or
# evCalc, although in the latter case we may need to provide is.ev=TRUE as well.

# Calculate and print statistics of posterior parameter PDFs given samp,
# assumed to have been sampled from the posterior.
# If samp has been sampled from the prior (i.e. is evCalc) then we must specify
# is.ev=TRUE, in which case importance sampling is used, but this is then unlikely
# to give accurate results!
# hfac is only used for calculating mean and sd if is.ev=TRUE.
# Statistics refer to the posterior PDF. So the mode is the MAP solution.
# The sd is the sqrt(variance) of the distribution, not of the mean.
# If Nrej is 0, use all the data. Otherwise, if it's a scalar, remove first Nrej samples,
#   and if it's a vector, remove those elements
# If thin>1, then this is used to thin out the samples prior to calculating statistics.
# Thinning is applied in blocks of Nthinblock: set this to Nwalker if emcee was used.
# k may be ignored by the direct user. k!=NA indicates that it the function been called
# by calc.post.stat.kfoldcv, in which case something is returned for use there,
# otherwise nothing returned by this function (just used to print).
# NOTE: is.ev=TRUE has not been tested
calc.post.stats <- function(samp, Nrej=0, thin=1, Nthinblock=1, hfac=NA, is.ev=FALSE, k=NA) {
  if(!identical(Nrej, 0)) {
    if(length(Nrej)==1) {
      samp <- samp[-(1:Nrej),]
    } else {
      samp <- samp[-(Nrej),]
    }
  }
  sel <- c(rep(FALSE,Nthinblock*(thin-1)),rep(TRUE,Nthinblock))
  resStats <- matrix(NA, nrow=ncol(samp)-2, ncol=3) # c(meanPar, modePar, sdPar) for each parameter
  if(is.na(k)) {
    cat("#    theta       mean        mode        sd\n")
  } else {
    cat(noquote(formatC(k, format="d", width=2)), "   theta       mean        mode        sd\n", sep="")
  }
  for(p in 3:ncol(samp)) {
    if(is.ev) { # if samp is the evidence then use importance sampling: weight is the likelihood
      hfac <- ifelse(is.na(hfac), -1*median(samp[,2]), hfac)
      impWeight <- 10^(samp[sel,2] + hfac)
      meanPar <- sum(impWeight*samp[sel,p])/sum(impWeight) 
      varPar  <- sum((impWeight*(samp[sel,p] - meanPar)^2))/sum(impWeight)
    } else {
      meanPar <- mean(samp[sel,p])
      varPar  <- var(samp[sel,p])
    }
    ind <- (1:nrow(samp))[sel] # one could argue that thinning is not needed for the mode, but I apply it anyway
    modePar <- samp[ind[which.max(samp[ind,1] + samp[ind,2])], p] 
    
#    word <- c( noquote(formatC(colnames(samp)[p], format="s", width=10)), noquote(formatC(meanPar, format="f", digits=5, width=11)), noquote(formatC(modePar, format="f", digits=5, width=11)), noquote(formatC(sqrt(varPar), format="f", digits=5, width=11)), "\n")
    word <- c( noquote(formatC(colnames(samp)[p], format="s", width=10)), noquote(formatC(meanPar, format="g", digits=4, width=11)), noquote(formatC(modePar, format="g", digits=4, width=11)), noquote(formatC(sqrt(varPar), format="g", digits=4, width=11)), "\n")
    cat(word)
    resStats[p-2,] <- c(meanPar, modePar, sqrt(varPar))
  }
  if(!is.na(k)) {
    return(resStats)
  }
}

# Does not take samp. Instead takes kfoldCV and calls calc.post.stats()
# for each partition. Note that the statistics calculated for partition k
# is for samples calculated using obsdata NOT in partition k, i.e. for
# the "training set" corresponding to partition k.
# Nrej and thin are applied to each partition.
# Thinning is applied in blocks of Nthinblock: set this to Nwalker if emcee was used.
calc.post.stats.kfoldcv <- function(kfoldCV, Nrej=0, thin=1, Nthinblock=1) {
  Ntheta <- dim(kfoldCV$multiPostSamp)[2] - 2 # number of parameters
  Npart  <- dim(kfoldCV$multiPostSamp)[3]     # number of partitions
  resStats <- array(NA, dim=c(Ntheta, 3, Npart))
  for(k in 1:Npart) {
    postSamp <- as.matrix(kfoldCV$multiPostSamp[,,k])
    colnames(postSamp) <- kfoldCV$postSampNames
    resStats[,,k] <- calc.post.stats(samp=postSamp, Nrej=Nrej, thin=thin, Nthinblock=Nthinblock, is.ev=FALSE, k=k)
  }
  cat("# Summary statistics across partitions\n")
  cat("#    theta     mean(mean)  sd(mean)       mean(mode)  sd(mode)       mean(sd)\n")
  for(p in 1:Ntheta) {
    word <- c( noquote(formatC(kfoldCV$postSampNames[p+2], format="s", width=10)), noquote(formatC(mean(resStats[p,1,]), format="g", digits=4, width=11)), noquote(formatC(sd(resStats[p,1,]), format="g", digits=4, width=11)), "  ", noquote(formatC(mean(resStats[p,2,]), format="g", digits=4, width=11)), noquote(formatC(sd(resStats[p,2,]), format="g", digits=4, width=11)), "  ", noquote(formatC(mean(resStats[p,3,]), format="g", digits=4, width=11)), "\n")
    cat(word)
  }
  cat("kfoldCV logLike = ", sum(kfoldCV$logLikePart), "\n")
}

# Re-calculate posterior-averaged log10 likelihood optionally using sample
# removal, thinning and/or an hfac.
# thin retains just every thin^th sample.
# If Nrej is 0, use all the data. Otherwise, if it's a scalar, remove first Nrej samples,
#   and if it's a vector, remove those elements
# If emcee was used then:
# - Nthinblock=Nwalker (thinning is applied in blocks of Nthinblock).
# - Nrej should be a (or contain multiples of) Nwalker
# Returns a scalar
recalc.like <- function(samp, Nrej=0, thin=1, Nthinblock=1, hfac=NA) {
  if(!identical(Nrej, 0)) {
    if(length(Nrej)==1) {
      samp <- samp[-(1:Nrej),]
    } else {
      samp <- samp[-(Nrej),]
    }
  }
  sel <- c(rep(FALSE,Nthinblock*(thin-1)),rep(TRUE,Nthinblock))
  logLike <- samp[sel,2]
  hfac <- ifelse(is.na(hfac), -1*median(logLike), hfac)
  return( log10(mean(10^(logLike+hfac))) - hfac ) 
}

# As recalc.like - see notes on that - but for K-fold CV likelihood.
# Returns 2-elements list:
# logLikePart, the log10 likelihood for each partition
# kfoldCVlogLike, the log10 K-fold CV likelihood (the sum of these).
recalc.kfoldcv.like <- function(kfoldCV, Nrej=0, thin=1, Nthinblock=1, hfac=NA) {
  if(!identical(Nrej, 0)) {
    if(length(Nrej)==1) {
      samp <- kfoldCV$multiPostSamp[-(1:Nrej),,]
    } else {
      samp <- kfoldCV$multiPostSamp[-(Nrej),,]
    }
  } else {
    samp <- kfoldCV$multiPostSamp
  }
  sel <- c(rep(FALSE,Nthinblock*(thin-1)),rep(TRUE,Nthinblock))
  logLikePart <- apply(samp[sel,,], 3, function(x) {   ########## DOUBLE CHECK THIS <<<<<<<<<<<<<<<<
    hfac <- ifelse(is.na(hfac), -1*median(x[,2]), hfac)
    log10(mean(10^(x[,2]+hfac))) - hfac
  } ) # x is 2D postSamp matrix for one partition. Returns log10 likelihood for each partition
  return( list(logLikePart=logLikePart, kfoldCVloglike=sum(logLikePart)) ) 
}

# For samp=evCalc or postSamp or kfoldCV$multiPostSamp[,,k]
# Find the parameter combination which has the largest posterior probability
# for parameter p (=column p+2 in samp) restricted to lie in the range prange.
# If p or prange is NA then this logically means find the global maximum,
# the MAP solution. This will be the same as the mode given by calc.post.stats(),
# Specifying a range is typically used to find the
# "second best solution" over a restricted parameter range.
# Note: this does not necessarily correspond to the value of parameter p at the
# second highest peak in the *marginal* posterior plotted in 1D for parameter p.
# If Nrej is 0, use all the data. Otherwise, if it's a scalar, remove first Nrej samples,
#   and if it's a vector, remove those elements
# The parameters are returned as a vector
find.peak.sol <- function(samp, Nrej=0, p=NA, prange=NA) {
  if(!identical(Nrej, 0)) {
    if(length(Nrej)==1) {
      samp <- samp[-(1:Nrej),]
    } else {
      samp <- samp[-(Nrej),]
    }
  }
  if(is.na(p) || any(is.na(prange)) ) {
    # cat("Finding MAP solution\n")
    peakpos <- which.max(samp[,1]+samp[,2])
  } else {
    ind <- which(samp[,p+2]>prange[1] & samp[,p+2]<prange[2])
    peakpos <- ind[which.max((samp[ind,1]+samp[ind,2]))]
  }
  #print(samp[peakpos,3:ncol(samp)])
  #return( as.list(samp[peakpos, 3:ncol(samp)] ))
  return( samp[peakpos, 3:ncol(samp)] )
}

# Does not take samp. Instead takes kfoldCV and calls find.peak.sol()
# for each partition.
find.peak.sol.kfoldcv <- function(kfoldCV, Nrej=0, p=NA, prange=NA) {
  x <- kfoldCV$postSampNames[3:length(kfoldCV$postSampNames)]
  cat(noquote( formatC(x, format="s", width=10, flag="") ), "\n")
  for(k in 1:dim(kfoldCV$multiPostSamp)[3]) {
    postSamp <- as.matrix(kfoldCV$multiPostSamp[,,k])
    colnames(postSamp) <- kfoldCV$postSampNames
    x <- find.peak.sol(samp=postSamp, Nrej=Nrej, p=p, prange=prange)
    cat(noquote( formatC(x, format="e", digits=3, width=10, flag="") ), "\n")
  }
}

# Only for samp=evCalc or postSamp.
# Calculate and print AIC, BIC by finding the maximum likelihood in the
# likelihoods in samp. Also return this for the No-model (for which AIC=BIC
# as it has no adjustable parameters).
# Really we have to find the maximum likelihood properly: finding it from
# samples of the posterior or (worse) prior is not correct.
# Nothing returned.
calc.xic <- function(samp, thetaFlag, obsdata, logLikeNomod) {
  lnLikeMax <- log(10)*max(samp[,2])  # natural logarithm
  Nparam <- length(thetaFlag[thetaFlag])
  aic <- -2*lnLikeMax + 2*Nparam
  bic <- -2*lnLikeMax + log(nrow(obsdata))*Nparam
  cat("AIC =", aic, "\n")
  cat("BIC =", bic, "\n")
  cat("AIC/BIC (No-model) =", -2*log(10)*logLikeNomod, "\n")
}



##### Plot MCMC posterior samples and 1D PDFs from postSamp or kfoldCV

# postSamp[,1:2] = c(logPrior, logLike). Other columns are the adjustable parameters.

# Plot MCMC trace and 1D density (i.e. unnormalized posterior PDF) for each of the adjustable
# parameters in postSamp which are specified by pcol. If passdat is a list, this is assumed to be
# kfoldCV, i.e. of format returned by kfold.cv(), from which postSamp is extracted and a plot produced
# for each partition. Otherwise we assume passdat=postSamp.
# Each 1D density is calculated by density estimation, and the mode of each is calculated and printed.
# Note that this is not the same as the global mode (MAP), which is the parameters of the single
# global maximum of the unsmoothed (non-binned) posterior.
# If working on kfoldCV, the mean and sd of the modes across all partitions is also calculated and printed.
# xlim specifies the parameter plotting (and mode estimation) limits for each parameter as
#   ordered in pcol, as a list, e.g. xlim=list( c(0,0.5), NULL, NULL, c(0,0.3) ).
#   NULL means it finds its own limits, and xmin=NULL finds own limits for all parameters.
#   xlim is applied after thinning but before density estimation, so that Ndense
#   is the number of samples *within the selected range*. As the mode just corresponds to the
#   largest of these sample, it will depend on Ndense.
# pbound specifies which of the parameter in postSamp should obey a periodic boundary condition
#   with limits given by xlim. It is specified as a logical vector ordered as in pcol. For those
#   elements which are true, density estimation is calculated by padding the ends of the data
#   vector with data from the other end. The kernel size is hard-wired into the code as fraciton
#   of the range specified in xlim. Set pbound=NULL or omit entirely to skip this for all parameters.
# Ndense is number of samples used to make the density estimate. Best set as a power of 2.
#   Should be large enough to get good resolution
# If Nrej is 0, use all the data. Otherwise, if it's a scalar, remove first Nrej samples,
#   and if it's a vector, remove those elements
# thin is the thinning factor, used only for calculating and plotting the posterior density.
# Thinning is applied in blocks of Nthinblock: set this to Nwalker if emcee was used.
# Ntrace is number of points used to plot trace (every n^th point is taken, n ~ length/Ntrace).
#   Note that if emcee has been used, these are selected from across all walkers.
#   Set to NA to plot all. (The thinning is not applied to the trace plot.)
# mfrow and plotsize specify plotsize. Plot is written to file fname (to screen if NA).
# Note that as postSamp provides logPrior and logLike, we could plot the posterior directly.
# However, this is very noise, so density estimation is preferred.
# If priorSamp is not NULL then it is assumed to contain samples from the priors for the adjustable
# parameters. It should be a 2D matrix (like postSamp) or a dataframe, with nrow as postSamp and
# the same parameter columns (in the same order), but without the first two columns (logPrior, logLike).
# myplot() does the plotting and returns the mode for each 1D, smoothed (density-estimated) PDF,
#   as the vector mode1dPar (which has ncol(postSamp) = Ntheta+2 elements)
plot.postSamp1d <- function(passdat, priorSamp=NULL, pcol=NA, xlim=NULL, pbound=NULL, Nrej=0, thin=1, 
                            Nthinblock=1, Ntrace=1e3, Ndense=2^12, fname=NA, mfrow=c(4,2), plotsize=c(6,4)) {
  
  myplot <- function(postSamp, pcol, thin=1, Nthinblock=1, priorSamp=NULL, xlim=NULL, pbound=NULL) {
    # do checks
    if(!is.null(xlim) && length(xlim)!=length(pcol)) {
      stop("xlim is not the same length as the number of parameters to plot")
    }
    if(!is.null(pbound) && length(pbound)!=length(pcol)) {
      stop("pbound is not the same length as the number of parameters to plot")
      if(length(pbound)!=length(xlim)) {
        stop("pbound selected, but is not the same length as xlim\n")
      }
    }
    mode1dPar <- c(rep(NA,length.out=ncol(postSamp))) # NA vector; elements pcol will be filled
    
    for(p in pcol) {
      Nsamp  <- nrow(postSamp)
      Ntrace <- ifelse(is.na(Ntrace), Nsamp, min(Ntrace, Nsamp))
      tracesel <- seq(from=1,to=Nsamp, length.out=Ntrace)
      plot(tracesel, postSamp[tracesel,p], type="l", ylim=xlim[[p-2]], xlab="iteration", ylab=colnames(postSamp)[p])
      # thinsel <- (1:Nsamp)[c(rep(FALSE,thin-1),TRUE)] # elements of postSamp after thinning
      thinsel <- (1:Nsamp)[c(rep(FALSE,Nthinblock*(thin-1)),rep(TRUE,Nthinblock))] # elements of postSamp after thinning
      if(is.null(xlim[[p-2]])) {
        densel <- thinsel
      } else {
        densel <- thinsel[ which(postSamp[thinsel,p] > xlim[[p-2]][1] & postSamp[thinsel,p] < xlim[[p-2]][2]) ] # xrange
      # The which statement gives those elements of thinsel for which postSamp satisfies the conditions.
      # As indices of thinsel, this then gives the elements of postSamp which satisfy the conditions.
      # Thus densel are those elements of postSamp which are thinned and *then* satisfy the range condition.
      # This order is important, because thinning applies to the whole range; the range is just for plotting
      }
      # calculate density either directly or by taking into account periodic boundary conditions
      if(!is.null(pbound) && pbound[p-2]) { # periodic boundary condition over specified xrange
        if(is.null(xlim) || is.null(xlim[[p-2]])) {
          stop("xlim must be defined when using pbound, but is not for parameter ", p-2)
        }
        x <- postSamp[densel,p]
        xup <- xlim[[p-2]][2]
        xlo <- xlim[[p-2]][1]
        hw <- (xup-xlo)/50
        fac <- 2
        # extend lower end by fac*hw from upper end, and upper end by fac*hw from lower end
        x <- c(x, x[which(x>xup-fac*hw)]-(xup-xlo), x[which(x<xlo+fac*hw)]+(xup-xlo))
        den <- density(x, n=Ndense, kernel="epanechnikov", width=hw, from=xlo, to=xup)
      } else {
        den <- density(postSamp[densel,p], n=Ndense, cut=0)
      }
      
      plot(den$x, den$y, type="l", xlim=xlim[[p-2]], xlab=colnames(postSamp)[p], ylab="density")
      mode1dPar[p] <- den$x[which.max(den$y)] # the 1D mode for this parameter
      # plot(postSamp[,p], 10^(postSamp[,1]+postSamp[,2]), type="h", xlab=colnames(postSamp)[p], ylab="Unnormalized posterior PDF")
      # plot prior is supplied
      if(!is.null(priorSamp)) {
        if(is.null(xlim[[p-2]])) {
          densel <- 1:nrow(priorSamp)
        } else {
          densel <- which( priorSamp[,p-2] > xlim[[p-2]][1] & priorSamp[,p-2] < xlim[[p-2]][2]) # xrange
        }
        den <- density(priorSamp[densel,p-2], n=Ndense, cut=0)
        lines(den$x, den$y, type="l", col="red")
      }
      
      mytit <- paste("Ndense=", Ndense, " Nrej=", Nrej, " thin=", thin, " Nthinblock=", Nthinblock, sep="")
      mtext(mytit, outer=TRUE)
    }
    return(mode1dPar)
    
  }
  
  if(!is.na(fname)) {
    pdf(fname, width=plotsize[1], height=plotsize[2])
  }
  par(mfrow=mfrow, mar=c(3.5,3.5,1.0,0.5), oma=c(1,1,3,1), mgp=c(2.0,0.6,0))

  if(is.list(passdat)) { # for kfoldCV
    mode1dPar <- array(NA, dim=dim(passdat$multiPostSamp)[2:3]) # mode1dPar[parameter, partition]
    if(any(is.na(pcol))) {pcol <- c(3:dim(passdat$multiPostSamp)[2])}
    cat("1D modes:\n")
    cat("  ", noquote(formatC(passdat$postSampNames[pcol], format="s", width=10)), "\n")
    for(k in 1:dim(passdat$multiPostSamp)[3]) {
      if(!identical(Nrej, 0)) {
        if(length(Nrej)==1) {
          postSamp <- as.matrix(passdat$multiPostSamp[-(1:Nrej),,k])
        } else {
          postSamp <- as.matrix(passdat$multiPostSamp[-(Nrej),,k])
        }
      } else {
        postSamp <- as.matrix(passdat$multiPostSamp[,,k])
      }
      colnames(postSamp) <- passdat$postSampNames
      mode1dPar[,k] <- myplot(postSamp, pcol, thin, Nthinblock, priorSamp, xlim, pbound)
      cat(noquote(formatC(k, format="d", width=2)), noquote(formatC(mode1dPar[pcol,k], format="g", digits=4, width=10)), "\n")
    }
    cat("# Mean and sd across partitions of the 1D modes\n")
    cat("#    theta     mean(mode)  sd(mode)\n")
    for(p in pcol) {
      word <- c( noquote(formatC(kfoldCV$postSampNames[p], format="s", width=10)), noquote(formatC(mean(mode1dPar[p,]), format="g", digits=4, width=11)), noquote(formatC(sd(mode1dPar[p,]), format="g", digits=4, width=11)), "\n")
      cat(word)
    }
  } else { # for postSamp
    if(any(is.na(pcol))) {pcol <- c(3:ncol(passdat))}
    if(!identical(Nrej, 0)) {
      if(length(Nrej)==1) {
        passdat <- passdat[-(1:Nrej),]
      } else {
        passdat <- passdat[-(Nrej),]
      }
    }
    mode1dPar <- myplot(passdat, pcol, thin, Nthinblock, priorSamp, xlim, pbound)
    cat("1D modes:\n", noquote(formatC(colnames(passdat)[pcol], format="s", width=10)), "\n")
    cat("", noquote(formatC(mode1dPar[pcol], format="g", digits=4, width=10)), "\n")
  }

  if(!is.na(fname)) {
    dev.off()
  }
  
}



##### Plot autocorrelation function of posterior samples for K-fold CV, with optional thinning

plot.chain.acf <- function(kfoldCV, Nrej=0, thin=1, Nthinblock=1, lag.max=NULL, fname=NA, plotsize=c(8,2)) {
  if(!identical(Nrej, 0)) {
    if(length(Nrej)==1) {
      samp <- kfoldCV$multiPostSamp[-(1:Nrej),,]
    } else {
      samp <- kfoldCV$multiPostSamp[-(Nrej),,]
    }
  } else {
    samp <- kfoldCV$multiPostSamp
  }
  Npart  <- dim(kfoldCV$multiPostSamp)[3]
  Ntheta <- dim(kfoldCV$multiPostSamp)[2]-2
  sel <- c(rep(FALSE,Nthinblock*(thin-1)),rep(TRUE,Nthinblock))
  if(!is.na(fname)) {
    pdf(fname, width=plotsize[1], height=plotsize[2])
  }
  par(mfrow=c(1,Ntheta), mar=c(3.0,3.0,3.0,0.5), oma=c(0.5,0.5,1.5,0.5), mgp=c(1.75,0.6,0))
  for(k in 1:Npart) {
    for(p in 1:Ntheta) {
      acf(samp[sel,p+2,k], lag.max=lag.max, cex.main=0.8, main="")
      mtext(kfoldCV$postSampNames[p+2])
    }
    mytit <- paste("thin=", thin, " partition=", k, sep="")
    mtext(mytit, outer=TRUE)
  }
  if(!is.na(fname)) {
    dev.off()
  }
}



##### Plot chains from parallel tempering

# Plot postSampAll: one panel per parameter, all chains in each parameter, cold chain in red.
# thetaFlag is passed to provide the variable names
plot.pt.chains <- function(postSampAll, thetaFlag, Ntrace=1e3, mfrow=c(2,3), plotsize=c(6,5), fname=NA) {
  Nsamp  <- dim(postSampAll)[1]
  Nchain <- dim(postSampAll)[2]
  Ntheta <- dim(postSampAll)[3]-2
  Ntrace <- ifelse(is.na(Ntrace), Nsamp, min(Ntrace, Nsamp))
  ind <- seq(from=1,to=Nsamp, length.out=Ntrace)
  if(!is.na(fname)) {
    pdf(fname, width=plotsize[1], height=plotsize[2])
  }
  par(mfrow=mfrow, mar=c(3.5,3.5,1.0,0.5), oma=c(1,1,3,1), mgp=c(2.0,0.6,0))
  for(p in 3:(Ntheta+2)) {
    plot(ind, postSampAll[ind,1,p], ylim=range(postSampAll[ind,,p]), type="n", xlab="iteration", ylab=names(which(thetaFlag))[p-2] )
    for(c in 1:Nchain) {
      lines(ind, postSampAll[ind,c,p], col=ifelse(c==Nchain,"red","black"))
    }
  }
  if(!is.na(fname)) {
    dev.off()
  }
}



##### Plot walker paths from emcee

# This is applied to either postSampAll of kfoldCV. If the latter then must specify
# partition number to plot ("part") and no. of walkers ("Nwalker").
# plot Ntrace samples over the iteration range xlim (if NA then plot all)
# thetaFlag is passed to provide the variable names
# Assumes order of dimensions is postSampAll[iteration, walker, theta]
# and kfoldCV$multiPostSamp[walker + Nwalker*(iteration-1), theta, partition]
plot.emcee.walkers <- function(passdat, part=1, Nwalker=NULL, thetaFlag, Ntrace=1e2, xlim=NULL, mfcol=c(5,2), plotsize=c(5,7), fname=NA) {

  if(is.list(passdat)) { # for kfoldCV
    Ncol  <- dim(passdat$multiPostSamp)[2]
    if(dim(passdat$multiPostSamp)[1] %% Nwalker != 0) {
      stop("First passdat dimension (", dim(passdat$multiPostSamp)[1], ") is not an integer multiple of Nwalker = ", Nwalker)
    }
    Nsamp <- dim(passdat$multiPostSamp)[1] %/% Nwalker
    if(part > dim(passdat$multiPostSamp)[3]) {
      stop("Partition " , part, " does not exist. Only ", dim(passdat$multiPostSamp)[3], " partitions")
    }
    # convert kfoldCV$multiPostSamp[walker + Nwalker*(iteration-1), theta, k]
    #      -> postSampAll[iteration, walker, theta] for k=part
    postSampAll <- aperm(array(passdat$multiPostSamp[,,part], dim=c(Nwalker, Nsamp, Ncol)), c(2,1,3))
  } else { # for postSampAll
    postSampAll <- passdat
  }
    
  Nsamp   <- dim(postSampAll)[1]
  Nwalker <- dim(postSampAll)[2]
  Ntheta  <- dim(postSampAll)[3]-2
  Ntrace <- ifelse(is.na(Ntrace), Nsamp, min(Ntrace, Nsamp))
  if(is.null(xlim)) {
    xlim <- 1:Nsamp
  }
  ind <- seq(from=min(xlim),to=max(xlim), length.out=Ntrace)
  if(!is.na(fname)) {
    pdf(fname, width=plotsize[1], height=plotsize[2])
  }
  par(mfcol=mfcol, mar=c(1,3.5,0,0.5), oma=c(5,1,3,1), mgp=c(2.0,0.6,0))
  panelcounter <- 0
  for(w in 1:Nwalker) {
    for(p in 3:(Ntheta+2)) {
      plot(ind,  postSampAll[ind,w,p], type="n", xaxt="n", ylab=names(which(thetaFlag))[p-2] )
      
      # could use this to achieve common yrange: ylim=range(postSampAll[ind,,p])
      lines(ind, postSampAll[ind,w,p])
      panelcounter <- panelcounter+1
      if(panelcounter %% mfcol[1] == 0) {
        axis(side=1, at=as.integer(seq(from=min(xlim), to=max(xlim), length.out=5)), outer=TRUE) 
        mtext(text="iteration", side=1, outer=FALSE, line=3)
      }      
      if(panelcounter %% mfcol[1] == 1) {
        mtext(text=paste("walker", w, " Ntrace", Ntrace), side=3, outer=FALSE, cex=0.8)
      }      
    } 
  }    
  if(!is.na(fname)) {
    dev.off()
  }
}



##### Plot autocorrelation function of each emcee walker, for single partition case

# thetaFlag is passed to provide the variable names
plot.walker.acf <- function(postSampAll, thetaFlag, Nrej=0, thin=1, lag.max=NULL, ylim=NULL, fname=NA, plotsize=c(8,2)) {
  Nwalker <- dim(postSampAll)[2]		
  Ntheta  <- dim(postSampAll)[3]-2
  if(!identical(Nrej, 0)) {
    if(length(Nrej)==1) {
      postSampAll <- postSampAll[-(1:Nrej),,]
    } else {
      postSampAll <- postSampAll[-(Nrej),,]
    }
  }
  sel <- c(rep(FALSE,thin-1),TRUE)
  if(!is.na(fname)) {
    pdf(fname, width=plotsize[1], height=plotsize[2])
  }
  par(mfrow=c(1,Ntheta), mar=c(3.0,3.0,3.0,0.5), oma=c(0.5,0.5,1.5,0.5), mgp=c(1.75,0.6,0))
  for(w in 1:Nwalker) {
    for(p in 3:(Ntheta+2)) {
      acf(postSampAll[sel,w,p], lag.max=lag.max, ylim=ylim, cex.main=0.8, main="")
      mtext(names(which(thetaFlag))[p-2])
    }
    mytit <- paste("walker", w, " thin", thin)
    mtext(mytit, outer=TRUE)
  }
  if(!is.na(fname)) {
    dev.off()
  }
}



##### Plot results of evidence calculation: plot prior and posterior PDFs in 1D and 2D

# evCalc[,1:2] = c(logPrior, logLike). Other columns are the adjustable parameters.
# Do not use these to plot postSamp, i.e. posterior samples. The plots will work but
# don't make much sense. 

# Plot 1D distributions of normalized prior and normalized posterior PDF for each parameter.
# Normalized means the area under the curves is 1.
# The prior PDF is just for that parameter, and is calculated from the distribution of the
# prior samples for that parameter (evCalc[,]) using kernel density estimation with density() 
# Npts points.
# Note that this is NOT the logPrior in evCalc[,1], which is the prior for all parameters.
# The posterior PDF is of course the posterior for all parameters, as it is calculated from
# the total prior and likelihood, but is then plotted w.r.t each parameter
# (an implicit marginalization). This is kernel smoothed and at Npts using a bandwidth equal to
# full range of the (prior) parameters divided by Npts. Note that postProb is not normalized,
# because it's just the posterior calculated at a set of samples from the prior.
# (Note also that we cannot do density estimation on a set of samples (x only) to plot the posterior,
# because the samples have not been drawn from the posterior. Instead we have (x,y) values.)
# hfac is the help factor to calculate evidence correctly in limit of small numbers.
# pcol specifies which parameters to plot as columns of evCalc (between 3 and ncol(evCalc) inclusive).
# xlim specifies the parameter plotting limits for each parameter as ordered in pcol, as a list, e.g.
#   xlim=list( c(0,0.5), NULL, NULL, c(0,0.3) ).
#   NULL means it finds its own limits, and xmin=NULL finds own limits for all parameters
# Npts is the number of equally-spaced points to use to *calculate* the histograms. The number
#  you can actually see in the plot depends on xlim.
# yzero=TRUE forces ymin=0, otherwise it's calculated from the data
plot.1dpost <- function(evCalc, hfac=NA, pcol, xlim=NULL, yzero=FALSE, Npts=500, fname="evCalcPost1D.pdf", mfrow=c(2,2), plotsize=c(6,6)) {
  if(is.na(hfac)) {hfac <- -1*median(evCalc[,2])}  
  if(length(pcol)<=0) {stop("Must specify at least one column to plot!")}
  modevidence <- mean(10^(evCalc[,2] + hfac))
  postProb <- 10^(evCalc[,1] + evCalc[,2] + hfac)/modevidence 
  pdf(fname, width=plotsize[1], height=plotsize[2])
  par(mfrow=mfrow, mar=c(3.5,2.5,1.0,0.5), oma=c(1,1,3,1), mgp=c(2.0,0.6,0))
  for(p in pcol) {
    bw <- diff(range(evCalc[,p]))/Npts
    trans <- ksmooth(evCalc[,p], postProb, kernel="normal", bandwidth=bw, n.points=Npts)
    trans$y[which(is.na(trans$y))] <- 0
    xdiff <- trans$x[2]-trans$x[1] # ksmooth must give uniform spacing!
    trans$y <- trans$y/sum(xdiff*trans$y)
    ymin <- ifelse(yzero, 0, min(trans$y))
    yrange <- c(ymin, max(trans$y))
    plot(trans$x, trans$y, xlab=colnames(evCalc)[p], xlim=xlim[[p-2]], ylim=yrange, ylab="", type="s")
    #trans <- hist(evCalc[,p], breaks=Npts, plot=FALSE)
    #lines(trans$mids, trans$density, xlab=colnames(evCalc)[p], type="s", col="red")
    den <- density(evCalc[,p], n=Npts)
    lines(den$x, den$y, type="l", col="red") 
    # Following would be the total prior, but is not so useful to plot
    #trans <- ksmooth(evCalc[,p], 10^(evCalc[,1]), kernel="normal", bandwidth=bw, n.points=Npts)
    #trans$y[which(is.na(trans$y))] <- 0
    #xdiff <- trans$x[2]-trans$x[1] # ksmooth must give uniform spacing!
    #trans$y <- trans$y/sum(xdiff*trans$y)
    #lines(trans$x, trans$y/sum(trans$y), xlab=colnames(evCalc)[p], type="s", col="green")
    if( ((p-2) %% prod(mfrow) == 1) || (prod(mfrow)==1) ) { # add title if at beginning of page
      title("Normalized posterior (black) and normalized prior (red) PDF", outer=TRUE)
    }
  }
  dev.off()
}

# Plot 2D distribution of unnormalized prior and unnormalized posterior PDF for any two parameters,
# for the column numbers pcol in evCalc.
# Note that here the prior plotted here is the *total* prior for *all* parameters (i.e. 10^logPrior)
# and NOT the prior for just the two parameters plotted. This is different from the 1D plots!
# This is a slightly strange thing to plot and shouldn't be falsely interpreted!
# The figure caption includes the maximum logPostProb and logLike, which can be more positive
# than the largest colour plotted, because the corresponds to the largest cell (bin)
# Requires library(fields)
plot.2dpost <- function(evCalc, hfac=NA, pcol, fname="evCalcPost2D.pdf", dynrange=50) {
  if(is.na(hfac)) {hfac <- -1*median(evCalc[,2])}
  if(length(pcol)!=2) {stop("Must specify only two columns to plot!")}
  pdf(fname, height=4, width=9)
  par(mfrow=c(1,2), mar=c(3.5,4.5,3,1), oma=c(1,0,1,2), mgp=c(2.0,0.6,0))
  # plot posterior
  modevidence <- mean(10^(evCalc[,2] + hfac))
  logPostProb <- evCalc[,1] + evCalc[,2] - log10(modevidence) + hfac
  map <- as.image(logPostProb, x=evCalc[,c(pcol[1],pcol[2])], nrow=100, ncol=100)
  zmax <- max(map$z, na.rm=TRUE)
  zrange <- c(zmax-dynrange, zmax)
  maxLrow <- which.max(evCalc[,2])
  word1 <- noquote(formatC(evCalc[maxLrow,2], format="f", digits=2, flag=""))
  word2 <- paste( noquote(formatC(evCalc[maxLrow,pcol[1]], format="g", digits=3, flag="")), noquote(formatC(evCalc[maxLrow,pcol[2]], format="g", digits=3, flag="")) )
  maxProw <- which.max(logPostProb)
  word3 <- noquote(formatC(logPostProb[maxProw], format="f", digits=2, flag=""))
  word4 <- paste( noquote(formatC(evCalc[maxProw,pcol[1]], format="g", digits=3, flag="")), noquote(formatC(evCalc[maxProw,pcol[2]], format="g", digits=3, flag="")) )
  mytit <- paste("Unnormalized posterior\n max(logLike) =", word1, "at pars =", word2, "\n max(logPost) =", word3, "at pars =", word4, sep=" ")
  image.plot(map, zlim=zrange, main=mytit, xlab=colnames(evCalc)[pcol[1]], ylab=colnames(evCalc)[pcol[2]], cex.main=0.8)
  # plot prior
  map <- as.image(evCalc[,1], x=evCalc[,c(pcol[1],pcol[2])], nrow=100, ncol=100)
  zmax <- max(map$z, na.rm=TRUE)
  zrange <- c(zmax-2, zmax)
  maxLrow <- which.max(evCalc[,1])
  word1 <- noquote(formatC(evCalc[maxLrow,1], format="f", digits=2, flag=""))
  word2 <- paste( noquote(formatC(evCalc[maxLrow,pcol[1]], format="g", digits=3, flag="")), noquote(formatC(evCalc[maxLrow,pcol[2]], format="g", digits=3, flag="")) )
  mytit <- paste("Unnormalized total prior\n max(logPrior) =", word1, "at pars =", word2, sep=" ")
  image.plot(map, zlim=zrange, main=mytit, xlab=colnames(evCalc)[pcol[1]], ylab=colnames(evCalc)[pcol[2]], cex.main=0.8)
  #
  dev.off()
}
  


##### Overplot user-defined model solution on top of data

# plot obsdata and then overplot a function of class TSMod1class with
# parameters set in the named list paramlist. Plot written to file fname if
# supplied, otherwise to the default device. yzero=TRUE forces the minium
# plotted value of the y-axis to be zero. paramlist should be a named list.
# It can be conveniently provided by find.peak.sol(), for example.
# If plotting a model involving FuncUniform then priorform must be provided
# as a character string within paramlist.
# If plotting a model involving FuncQuasiSinusoid then fptimes must be provided
# as a numerical vector within paramlist.
# Requires library(gplots)
oplot.model.data <- function(obsdata, TSMod1class, fname=NA, yzero=FALSE, paramlist) {
  if(!is.na(fname)) {
    pdf(fname, width=6, height=3)
  }
  #
  trange <- range(c(obsdata[,1]+obsdata[,2], obsdata[,1]-obsdata[,2]))
  TSMod1 <- new(TSMod1class)
  # The following may define unused parameters, but these have no effect
  deflist <- c(priorform=NA, fptimes=NA, offset=NA, sigamp=NA, lambda=NA, tzero=NA, sinamp=NA, freq=NA, phase=NA, fp1=NA, fp2=NA, fp3=NA, fp4=NA, slope=NA, tmid=NA, sd=NA)
  paramlist <- c(paramlist, deflist) # paramlist takes priority over deflist when accessing
  if(.hasSlot(TSMod1, "priorform")) {
    TSMod1@priorform   <- paramlist$priorform
  }
  if(.hasSlot(TSMod1, "fptimes")) {
    TSMod1@fptimes     <- paramlist$fptimes
  }
  TSMod1@theta1$offset <- paramlist$offset
  TSMod1@theta1$sigamp <- paramlist$sigamp
  TSMod1@theta1$lambda <- paramlist$lambda
  TSMod1@theta1$tzero  <- paramlist$tzero
  TSMod1@theta1$sinamp <- paramlist$sinamp
  TSMod1@theta1$freq   <- paramlist$freq
  TSMod1@theta1$phase  <- paramlist$phase
  TSMod1@theta1$fp1    <- paramlist$fp1
  TSMod1@theta1$fp2    <- paramlist$fp2
  TSMod1@theta1$fp3    <- paramlist$fp3
  TSMod1@theta1$fp4    <- paramlist$fp4
  TSMod1@theta1$slope  <- paramlist$slope
  TSMod1@theta1$tmid   <- paramlist$tmid
  TSMod1@t <- seq(from=trange[1], to=trange[2], length.out=10000)
  ymod <- eval.mod(TSMod1)
  #
  par(mgp=c(2.0,0.8,0), mar=c(3.5,4.0,2.4,1.0), cex.main=0.8)
  ymin <- ifelse(yzero, 0, min(c(ymod, obsdata[,3]-obsdata[,4])))
  yrange <- c(ymin, max(c(ymod, obsdata[,3]+obsdata[,4])))
  plot(obsdata[,1], obsdata[,3], xlim=trange, ylim=yrange,  pch=18, xlab="measured time s", ylab="measured signal y")
  if(any(obsdata[,2]!=0)) {
    plotCI(obsdata[,1], obsdata[,3], uiw=obsdata[,2], err="x", type="n", gap=0, sfrac=0.01, add=TRUE)
  }
  if(any(obsdata[,4]!=0)) {
    plotCI(obsdata[,1], obsdata[,3], uiw=obsdata[,4], err="y", type="n", gap=0, sfrac=0.01, add=TRUE)
  }
  lines(TSMod1@t, ymod, col="red")
  if(!is.na(paramlist$sd)) {
    lines(TSMod1@t, ymod-paramlist$sd, col="red", lty=2)	
    lines(TSMod1@t, ymod+paramlist$sd, col="red", lty=2)	
  }
  # This is the only way I could find to get mytit to print on one line! 
  ind <- !is.na(TSMod1@theta1)
  mytit <- paste(names(TSMod1@theta1[ind]), rep("="), noquote(formatC(unlist(TSMod1@theta1[ind]), format="f", width=6)), sep="")
  if(!is.na(paramlist$sd)) {
    mytit <- c(mytit, "theta2sd=", noquote(formatC(paramlist$sd, format="f", width=4)), sep="")
  } 
  cat(mytit, file="xyz")
  mytit <- readLines("xyz", warn=FALSE)
  system("rm -f xyz")
  title(main=mytit, cex.main=0.6)
  #
  if(!is.na(fname)) {
    dev.off()
  }
}

# For TSMod2=ProbOUprocess or ProbWienerprocess
# plot obsdata and then overplot parameter of prior PDF of process variable, zMean ?? sqrt(zVar),
# at each point. The latter are calculated given the OUprocess parameters in paramlist
# and the obsdata. If relax is missing, then we use the Wiener process.
# Data are plotted in black, preductions in red.
# If plotResid=TRUE, instead plot data-prediction, with combined y error bars and no t error bars.
# If specified, predictions for events in selInd are (over)plotted in green (ignored if plotResid=TRUE).
# Plot written to file fname if supplied, otherwise to the default device. 
# yzero=TRUE forces the minium plotted value of the y-axis to be zero (ignored if plotResid=TRUE).
# trange=c(tmin,tmax) can be supplied, otherwise (if it's NULL) is calculated
# from the data.
# Requires library(gplots)
oplot.OUprocess.data <- function(obsdata, selInd=NULL, fname=NA, yzero=FALSE, plotResid=FALSE, trange=NULL, paramlist) {

  if(!is.na(fname)) {
    pdf(fname, width=6, height=3.2)
  }
  if(is.null(trange)) {
    trange <- range(c(obsdata[,1]+obsdata[,2], obsdata[,1]-obsdata[,2]))
  }

  if(is.element("relax", names(paramlist))) {
    TSMod2 <- new("ProbOUprocess")
  } else {
    TSMod2 <- new("ProbWienerprocess")
  }
  deflist <- c(relax=NA, diffcon=NA, offset=NA, zstartmean=NA, zstartsd=NA)
  paramlist <- c(paramlist, deflist) # paramlist takes priority over deflist when accessing
  TSMod2@theta2$relax      <- paramlist$relax
  TSMod2@theta2$diffcon    <- paramlist$diffcon
  TSMod2@theta2$offset     <- paramlist$offset
  TSMod2@theta2$zstartmean <- paramlist$zstartmean
  TSMod2@theta2$zstartsd   <- paramlist$zstartsd
  procParam <- eval.mod(TSMod2, obsdata)
  
  par(mfrow=c(1,1), mgp=c(2.0,0.8,0), mar=c(3.5,4.0,1.0,1.0), oma=c(0,0,2.0,0), cex.main=0.8)
  # To plot without captions use this:
  # par(mgp=c(2.0,0.8,0), mar=c(3.5,4.0,1.0,1.0), oma=c(0,0,0.0,0), cex.main=0.8)
  if(plotResid) {
    yval <- obsdata[,3] - procParam$zMean
    ysd  <- sqrt(procParam$zVar+obsdata[,4]^2)
    yrange <- c(min(yval-ysd), max(yval+ysd))
    plot(obsdata[,1], yval, xlim=trange, ylim=yrange,  pch=18, xlab="time", ylab="signal", cex=0.8)
    plotCI(obsdata[,1], yval, uiw=obsdata[,4], err="y", type="n", gap=0, sfrac=0.01, add=TRUE)
    abline(h=0, lwd=0.5)
  } else {
    ymin <- ifelse(yzero, 0, min(c(procParam$zMean-sqrt(procParam$zVar), obsdata[,3]-obsdata[,4])))
    yrange <- c(ymin, max(c(procParam$zMean+sqrt(procParam$zVar), obsdata[,3]+obsdata[,4])))
    plot(obsdata[,1], obsdata[,3], xlim=trange, ylim=yrange,  pch=18, xlab="time", ylab="signal", cex=0.8)
    plotCI(obsdata[,1], obsdata[,3], uiw=obsdata[,2], err="x", type="n", gap=0, sfrac=0.01, add=TRUE)
    plotCI(obsdata[,1], obsdata[,3], uiw=obsdata[,4], err="y", type="n", gap=0, sfrac=0.01, add=TRUE)
    points(obsdata[,1], procParam$zMean, pch=8, cex=0.5, col="red")
    plotCI(obsdata[,1], procParam$zMean, uiw=sqrt(procParam$zVar), err="y", type="n", col="red", gap=0, sfrac=0.01, add=TRUE)
    if(!is.null(selInd)) {
      if(!all(is.element(selInd, 1:nrow(obsdata)))) {stop("selInd contains indices not in 1:nrow(obsdata)")}
      points(obsdata[selInd,1], procParam$zMean[selInd], pch=8, cex=0.5, col="green")
      plotCI(obsdata[selInd,1], procParam$zMean[selInd], uiw=sqrt(procParam$zVar[selInd]), err="y", type="n", col="green", gap=0, sfrac=0.01, add=TRUE)
    }
  }
  # This is the only way I could find to get mytit to print on one line! 
  ind <- !is.na(TSMod2@theta2)
  mytit <- paste(names(TSMod2@theta2[ind]), rep("="), noquote(formatC(unlist(TSMod2@theta2[ind]), format="g", width=3)), sep="")
  cat(mytit, file="xyz")
  mytit <- readLines("xyz", warn=FALSE)
  system("rm -f xyz")
  title(mytit, outer=TRUE)
  if(plotResid) {
    mtext("residual, measured-model (y-z)", cex=0.8, padj=-1)
  } else {
    mtext("measured (y) = black, model prior (z) = red/green", cex=0.8, padj=-1)
  }
  
  if(!is.na(fname)) {
    dev.off()
  }
  
}
